/***************************************************************
 * config.h
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef _CONFIG_H_
#define _CONFIG_H_

#ifndef USE_SPU
#define USE_SPU	0
#endif

#if FASTEST
#define INLINE	inline
#else
#define INLINE
#endif

#if defined(__SPU__) || defined(__CYGWIN__)
#define HAS_FLOCK	0
#else
#define HAS_FLOCK	1
#endif

#if USE_SPU
#define GG_MAX_THREAD	6
#elif defined(__PPU__)
#define GG_MAX_THREAD	1
#else
#define GG_MAX_THREAD	16	/* Never set to 1 */
#endif

#if !defined(__SPU__) && GG_MAX_THREAD > 1
#define TLS	__thread
#else
#define TLS
#endif

#define CS_ADDR		"192.168.150.255"
#define CS_PORT		28124

#define GG_MAX_BOARD	9
#define USE_DYNAMIC_ALLOCATE	1

#if GG_MAX_BOARD <= 11
#define SMALLER_BOARD	1
#else
#define SMALLER_BOARD	0
#endif

#if SMALLER_BOARD
 #if USE_DYNAMIC_ALLOCATE
//  #define GG_MAX_PLAYOUT	1700000		/* For 8 GiB of RAM */
//  #define GG_MAX_PLAYOUT	850000		/* For 4 GiB of RAM */
  #define GG_MAX_PLAYOUT	10000		/* For CEDEC */
 #else
  #define GG_MAX_PLAYOUT	530000
 #endif
#else
 #if USE_DYNAMIC_ALLOCATE
  #define GG_MAX_PLAYOUT	320000
//  #define GG_MAX_PLAYOUT	160000
 #else
  #define GG_MAX_PLAYOUT	123000
 #endif
#endif

/********************************/
#define USE_EX_DELAY		0
#define MAX_EX_DELAY		100

#define USE_INTERNAL_PLAYOUT	1

/********************************/
#define PRINT_STATISTICS	0

#define PROFILING		0

#define USE_INDIVIDUAL_FACTOR	0

#define USE_ARC_CACHE		1

#define ADAPT_JAPANESE		1
#define JAPANESE_DONTPASS	(GG_MAX_MOVENUM / 10)

#define ADD_BLOCK		1

#define USE_PATTERN_TABLE	1
#define GENERATE_PATTERN	0

#define PRIOR_KO_TSUGI		0
#define CAPTURE_IS_INTEREST	0
#define PROHIBIT_UTTEGAESHI	0
#define PRIOR_UTTEGAESHI	0

#define USE_GC			0	/* Not implemented */
#define USE_HASH		0	/* Not implemented */

/* Used in mc.c */
/* Mode. Set one of the following flags to 1. */
#define PURE_RANDOM		0
#define MOGO			1

#define DETERMINISTIC		1

/* Special modes for Ko threats.
 * If the turn is immediate after a Ko capture then select
 * move as:
 *   case 0: usual (prune, or MOGO style),
 *   case 1: uniform randomly,
 *   case 2: uniformly from the union of steps 1, 2 and 3,
 *   case 3: uniformly from the union of steps 1 and 2.
 */
#define KO_THREAT		0

#define USE_RAVE		1
#define USE_RAVE_UPTO		0
#define USE_UCB1_TUNED		0
#define USE_EMPTY_LIST		1
#define USE_SPIRAL		1
#define MC_UNIQUE		1

/*******************************/

#if SMALLER_BOARD

#define PROPER_UCT	0
#define NO_VARIANT	1

#define TIME_MARGIN	30
#define MOVE_MARGIN	0

#define USE_ADJACENT_REN	0	/* Not implemented */
#define ADD_33			0	/* = # of add */
#define ADD_21			0	/* = # of add */
#define ADD_CAPTURE_SIZE	0	/* = Size of ren */
#define ADD_CAPTURE		0	/* = # of add */
#define ADD_NAKADE		0

/* BLOCK_MODE
 *  0: No blockmode at all
 *  1: Cfg
 *  2: Openning
 *  3: Cfg with openning (standard) */
#define BLOCK_MODE		3
#define BLOCK_DISTANCE		2
#define BLOCK_EXCLUDE		0	/* First XX moves are openning */
#define BLOCK_EXCLUDE_KO	1

#define USE_OFFSET		0
#define MAX_OFFSET		3

#define USE_WINNING_RATE	1
#define VISITS_ALLOWED		0.5
#define USE_LCB_CUT		1
#define LCB_FACTOR		2  /* 1=68.3%, 2=95.4%, 2.58=99%, 3=99.7% */
#define LCB_THRESHOLD		100000

#define ADD_PASS_THRESHOLD	0

#define USE_PRIOR		0
#define PRIOR_LIBERTIES		0
#define M_PRIOR			50
#define Q_PASS			0.1
#define LOW_PRIOR_SELF_ATARI	0
#define Q_SELF_ATARI		0.1

#define USE_CAPTURED		2
#define CAPTURED_THRESHOLD	(mc_board_size * 3 / 2)

#define USE_LADDER_MODE		0

#define USE_ORACLE		0
#define N_ORACLE		20

#else	/* SMALLER_BOARD (ie, LAGER BOARD) */

#define PROPER_UCT	0
#define NO_VARIANT	0

#define TIME_MARGIN	40
#define MOVE_MARGIN	0

#define USE_ADJACENT_REN	0
#define ADD_33			0	/* = # of add */
#define ADD_21			0	/* = # of add */
#define ADD_CAPTURE_SIZE	0	/* = Size of ren */
#define ADD_CAPTURE		0	/* = # of add */
#define ADD_NAKADE		0

/* BLOCK_MODE
 *  0: No blockmode at all
 *  1: Cfg
 *  2: Openning
 *  3: Cfg with openning (standard) */
#define BLOCK_MODE		3
#define BLOCK_DISTANCE		2	/* = Distance */
#define BLOCK_EXCLUDE		4	/* First N moves don't use block */
#define BLOCK_EXCLUDE_KO	1

#define USE_OFFSET		0
#define MAX_OFFSET		10

#define USE_WINNING_RATE	1
#define VISITS_ALLOWED		0.5
#define USE_LCB_CUT		1
#define LCB_FACTOR		2  /* 1=68.3%, 2=95.4%, 2.58=99%, 3=99.7% */
#define LCB_THRESHOLD		5000

#define ADD_PASS_THRESHOLD	1

#define USE_PRIOR		1
#define PRIOR_LIBERTIES		0
#define M_PRIOR			50
#define Q_PASS			0.01
#define LOW_PRIOR_SELF_ATARI	1
#define Q_SELF_ATARI		0.1

#define USE_CAPTURED		2
#define CAPTURED_THRESHOLD	(mc_board_size * 3 / 2)

#define USE_LADDER_MODE		0

#define USE_ORACLE		0
#define N_ORACLE		20

#endif	/* SMALLER_BOARD */

#endif /* _CONFIG_H_ */
