/***************************************************************
 * genmove.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "gggo.h"
#include "global.h"
#include "play.h"
#include "mc.h"
#include "queue.h"
#include "uct.h"
#include "cs.h"
#if USE_OFFSET
#include <math.h>
#endif

#if PROFILING
#include "timebase.h"
#define GG_N_PROFILE	7
static uint32_t profile[GG_N_PROFILE][GG_N_CLIENT_SHARE + 1];
#define PROFILE_VAR(var)	uint32_t var;
#define PROFILE_CLEAR	\
  for(int i = 0; i < GG_N_PROFILE; i += 1)	\
    for(int j = 0; j < GG_N_CLIENT_SHARE + 1; j += 1)	\
      profile[i][j] = 0;
#define PROFILE_START(var)	var = START_CLOCK();
#define PROFILE_RECORD(var, n, id)	profile[n][id] += DELTA_CLOCK(var);
#define PROFILE_RECORD_AND_START(var, n, id)	\
      PROFILE_RECORD(var, n, id)	\
      PROFILE_START(var)
#else
#define PROFILE_VAR(var)
#define PROFILE_CLEAR
#define PROFILE_START(var)
#define PROFILE_RECORD(var, n, id)
#define PROFILE_RECORD_AND_START(var, n, id)
#endif

/****************
 *	SFMT	*
 ****************/
#define init_gen_rand	gg_init_gen_rand
#define genrand_real2	gg_random
#define gen_rand32	gg_rand32
#define gen_rand64	gg_rand64

// #define get_idstring	gg_get_idstring
#define get_min_array_size32	gg_get_min_array_size32
#define get_min_array_size64	gg_get_min_array_size64
#define fill_array32	gg_fill_array32
#define fill_array64	gg_fill_array64
#define init_by_array	gg_init_by_array

#include "SFMT.h"

/* in mc.c */
extern void	mc_init_gen_rand(uint32_t seed);

/* in main.c */
extern ggTime 	gg_colored_left(ggColor color);

/* Provides */
void	gg_init_playout(void);
int	gg_finish_playout(void);
ggMove	gg_do_genmove(ggColor c);

/*********************************************/

static const char *
gg_ordinal_suffix(int i) {
  const char *suffices[] = {"th", "st", "nd", "rd"};
  return suffices[(1 <= i && i <= 3) ? i : 0];
}

/************************************************
 *	Playout					*
 ************************************************/
static uint32_t	gg_total_playout;

int
gg_finish_playout(void) {
#if 0
  int i, rc;
#endif

if(0) gprintf("--> %s()\n", __func__);
#if PRINT_STATISTICS
  fclose(gg_report);
#endif
#if 0
  for(i = 0; i < GG_N_CLIENT_SHARE; i += 1) {
    gg_push(CS_EXIT, cs_client[i].command);
  }
if(0) gprintf("Sent 'CS_EXIT' --- %s.\n", __func__);
  cs_usleep(100);
  for(i = 0; i < GG_N_CLIENT_SHARE; i += 1) {
    rc = pthread_join(cs_client[i].thread, NULL);
    if(rc != 0) {
      perror(__func__);
    }
  }
#endif
if(9) gprintf("<-- %s total_playout=%d\n", __func__, gg_total_playout);
  return gg_total_playout;
}

#if PROFILING
/* Print out profiled data.
 */
static void
gg_print_profile(uint32_t profile[GG_N_PROFILE][GG_N_CLIENT_SHARE + 1]) {

  for(int i = 0; i < GG_N_CLIENT_SHARE + 1; i += 1) {
    gprintf("Profile[%d]", i);
    for(int j = 0; j < GG_N_PROFILE; j += 1) {
      gprintf(" %.2fs", SECOND(profile[j][i]));
    }
    gputc('\n');
  }
  gprintf("Average:");
  for(int j = 0; j < GG_N_PROFILE; j += 1) {
    float t = 0;
    int percent;
    for(int i = 1; i < GG_N_CLIENT_SHARE + 1; i += 1) {
      t += SECOND(profile[j][i]);
    }
    t /= (float)GG_N_CLIENT_SHARE;
    percent = roundf(1e2 * t / SECOND(profile[0][0]));
    gprintf(" %.2fs(%d%%)", t, percent);
  }
  gputc('\n');

#if PRINT_STATISTICS
  for(int i = 0; i < GG_N_CLIENT_SHARE + 1; i += 1) {
    fprintf(gg_report, "Profile[%d]", i);
    for(int j = 0; j < GG_N_PROFILE; j += 1) {
      fprintf(gg_report, " %.2fs", SECOND(profile[j][i]));
    }
    fputc('\n', gg_report);
  }
  fprintf(gg_report, "Average:");
  for(int j = 0; j < GG_N_PROFILE; j += 1) {
    float t = 0;
    int percent;
    for(int i = 1; i < GG_N_CLIENT_SHARE + 1; i += 1) {
      t += SECOND(profile[j][i]);
    }
    t /= (float)GG_N_CLIENT_SHARE;
    percent = roundf(1e2 * t / SECOND(profile[0][0]));
    fprintf(gg_report, " %.2fs(%d%%)", t, percent);
  }
  fputc('\n', gg_report);
  fflush(gg_report);
#endif
  return;
}
#endif

static int
gg_mostvisitedarc(ggUctNode *node, int exclude) {
  float max;
  int bestarc, arc;

  assert(node != NULL);
  assert(node->n_arc > 0);
  bestarc = -1;
  max = -1;
  for(arc = 0; arc < node->n_arc; arc += 1) {
    if(arc == exclude)	continue;
    if(node->arc[arc].visits > max) {
      max = node->arc[arc].visits;
      bestarc = arc;
    }
  }
  return bestarc;
}

/* Return most probable arc.
 * Usually it's most visited arc but
 * check best valued arc as well.
 * */
static int
gg_winning_arc(ggUctNode *root, int exclude) {
  int bestarc;
#if USE_WINNING_RATE
  float maxvisit, v;
  int secondarc;
#endif

  bestarc = gg_mostvisitedarc(root, exclude);
#if USE_WINNING_RATE
  maxvisit = root->arc[bestarc].visits;
  v = root->arc[bestarc].value;
  secondarc = gg_mostvisitedarc(root, bestarc);
  if(secondarc > 0 && secondarc != exclude) {
    if(root->arc[secondarc].value > v &&
	 root->arc[secondarc].visits >= VISITS_ALLOWED * maxvisit) {
      bestarc = secondarc;
    }
  }
#endif
  return bestarc;
}

static void
gg_report_pv_1(FILE *report, ggUctNode *root, int n_repeat, float v) {
  ggUctNode *p;
  int bestarc, arc;
  float var;
  float visit;

  bestarc = gg_winning_arc(root, -1);
  if(bestarc >= 0) {
    var = root->arc[bestarc].ucb - root->arc[bestarc].value;
    visit = root->arc[bestarc].visits;
    gfprintf(report, "%2d%s %4.2f+%5.2f%%", n_repeat,
	    gg_ordinal_suffix(n_repeat), 100 * v, 100 * var);
    for(p = root; p != NULL; p = p->arc[arc].node) {
      if(p->position_sp->gameend)	break;
      arc = gg_winning_arc(p, -1);
      if(arc < 0)	break;
      gfprintf(report, " %1m", p->arc[arc].move);
    }
    fprintf(report, " n=%.0f", visit);
    arc = gg_winning_arc(root, bestarc);
    if(arc >= 0) {
      gfprintf(report, " (%1m=%4.2f+%5.2f%%,%.0f)", root->arc[arc].move,
			100 * root->arc[arc].value,
			100 * (root->arc[arc].ucb - root->arc[arc].value),
			root->arc[arc].visits);
    }
    gfprintf(report, "\n");
  } else {
    fprintf(report, "%2d%s %4.2f%%\n", n_repeat,
	      gg_ordinal_suffix(n_repeat), 100 * v);
  }
  return;
}

static void
gg_report_pv(ggUctNode *root, int n_repeat, float v) {
  assert(root != NULL);
  gg_report_pv_1(stderr, root, n_repeat, v);
#if PRINT_STATISTICS
  gg_report_pv_1(gg_report, root, n_repeat, v);
#endif
  return;
}

void
gg_init_playout(void) {
if(0) gprintf("--> %s()\n", __func__);
#if PRINT_STATISTICS
  gg_report = fopen("report", "a");
#endif
  gg_total_playout = 0;
  cs_init_client();
if(0) gprintf("<-- %s()\n", __func__);
  return;
}

#if USE_ORACLE
extern ggMove	gg_str2cord(const char *s, char **s0);
static void
gg_do_oracle(ggUctNode *root) {
  ggMove move;
  int i;
  float v, s;
  char *p;
  char oracle[GG_ORACLE_LINESIZE];

if(2) gprintf("--> %s\n", __func__);
  gg_put_oracle("gg_genmove %C", root->color);
  gg_get_oracle(&oracle[0]);
  gg_put_oracle("top_moves");
  p = gg_get_oracle(&oracle[0]);
  s = 0.5 / (float)(gg_board_size * gg_board_size);
  do {
    move = gg_str2cord(p, &p);
    if(move == 0)	break;
    v = 0.5 + s * strtof(p, &p);
    for(i = 0; i < root->n_arc; i += 1) {
      if(root->arc[i].move == move) {
#if N_ORACLE > 0
	root->arc[i].visits = (float)N_ORACLE;
	root->arc[i].value = v;
	root->arc[i].value2 = v * v;
	root->visits += (float)N_ORACLE;
if(2) gprintf("%1m=%.1f ", move, v);
#endif
	break;
      }
    }
  } while(true);
if(0) gputc('\n');

  for(i = 0; i < root->n_arc; i += 1) {
    if(root->arc[i].node != NULL) {
      root->arc[i].ucb = uct_ucb(root, i);
    }
if(2) gprintf("%1m=%.1f/%d ", root->arc[i].move, root->arc[i].ucb,
					      root->arc[i].visits);
  }
if(2) gputc('\n');
if(2) gprintf("<-- %s\n", __func__);
  return;
}
#endif

extern inline int	gg_n_empty_point(void);
extern inline int
gg_n_empty_point(void) {
  int n = 0;
  ggPoint p;
#if USE_EMPTY_LIST
  for(p = gg_empty_list[0]; p != 0; p = gg_empty_list[p]) {
    n += 1;
  }
#else
  GG_FOR_ALL_POINT(p) {
    if(gg_is_empty(p))	n += 1;
  }
#endif
  return n;
}

static int
gg_n_moves_left(ggColor color) {
  int n;
  ggMove move;

  n = 0;
  GG_FOR_ALL_EMPTY_POINT(move) {
    if(gg_is_illegal(move, color)) continue;
    n += 1;
  }
  return n;
}

#include <math.h>
static inline float
ucb_var(ggUctNode *p, int arc) {
  return sqrtf(logf(p->visits) / p->arc[arc].visits);
}

static timer_t timerid;
static struct sigevent timerevent;
static struct itimerspec itval;
static struct sigaction sigact;

static void
cs_timeup_handler(int signum) {
  UNUSED(signum);
if(0) fprintf(stderr, "--> %s(%d)\n", __func__, signum);
  cs_timeup = true;
  return;
}

static float gg_prev_v = 0.5;

/* do_genmove(ggColor c)
 * Called from gg_genmove_thread.
 */
ggMove
gg_do_genmove(ggColor c) {
  ggUctNode *root;
  float v;
  float maxvisit;
  int bestarc;
  ggMove bestmove;
  ggTime left, t;
  float moves;
  int repeat = 0;

if(0) gprintf("--> %s(%C)\n", __func__, c);

#if ADAPT_JAPANESE == 1
  /* Check if opponent passed. */
  /* TODO Check if I'm winning. */
  if(gg_previous_move == GG_PASS_MOVE && gg_movenum >= JAPANESE_DONTPASS) {
    bestmove = GG_PASS_MOVE;
    goto onlyone;
  }
#endif

  if(gg_in_joseki && gg_use_joseki) {
    bestmove = gg_genjoseki(gg_movenum);
    if(gg_is_onboard(bestmove))	goto onlyone;
    gg_in_joseki = false;
  }
#if 0
  if(gg_colored_left(c) < 1.5)	{
    goto play_pass;
  }
#endif

  /* Initialize.
   */
if(0) gprintf("Start initialize -- %s\n", __func__);

  /* Setup root node. */
  uct_init_tree();
  gg_generation += 1;	/* TODO if gc is implemented, move this */
  root = uct_create_node(c, gg_push_position(), NULL, 0);

  assert(root->n_arc > 0);
  if(root->n_arc == 1) {
    /* If only one move is allowed pick it up without search.
     * This includes end of the game. */
    bestmove = root->arc[0].move;
    goto onlyone;
  }

#if USE_ORACLE
  gg_do_oracle(root);	/* Setup initial ucb of all moves at root. */
#endif

/* Do playout.
 */
  left = gg_colored_left(c);
  if((t = gg_time_per_move) == 0) {
    if(left > TIME_MARGIN) {
      moves = gg_n_moves_left(c) - MOVE_MARGIN;
#if 0
      t = 4 * left / moves;
#elif SMALLER_BOARD
      t = 2 * left /
      ((float)SQR(GG_MAX_BOARD) * powf(moves / (float)SQR(GG_MAX_BOARD), 4.0));
#else
      t = 2 * left /
      ((float)SQR(GG_MAX_BOARD) * powf(moves / (float)SQR(GG_MAX_BOARD), 2.5));
#endif
#if 1 && SMALLER_BOARD
      if(0.45 <= gg_prev_v && gg_prev_v <= 0.55)	t *= 1.8;
      else if(gg_prev_v < 0.6)				t *= 1.5;
#else
      if(0.45 <= gg_prev_v && gg_prev_v <= 0.55)	t *= 1.5;
      else 						t *= 1.2;
#endif
    }
    t = MAX(0.04, t);
    t = MIN(72, t);
  }
if(2) fprintf(stderr, "Start playout left=%.2fs, empties=%3d, t=%.2fs\n",
		left, gg_n_moves_left(c), t);

  cs_create_timer(&timerevent, &timerid, &sigact, &cs_timeup_handler, SIGUSR1);
  cs_timeup = false;
  cs_set_timer(t, &itval, timerid);

  do {
    gg_total_playout += cs_do_playout(root, 5000);
    bestarc = (ggArc)gg_winning_arc(root, -1);
    assert(bestarc < root->n_arc);
    gg_prev_v = root->arc[bestarc].value;
    gg_report_pv(root, ++repeat, gg_prev_v);
    if(gg_prev_v >= 0.90)	break;
#if USE_LCB_CUT
    if(gg_prev_v > 0.5 && root->arc[bestarc].visits >= LCB_THRESHOLD) {
      float upper, lower;
      lower = root->arc[bestarc].value - LCB_FACTOR * ucb_var(root, bestarc);
      for(int arc = 0; arc < root->n_arc; arc += 1) {
	if(arc == bestarc)	continue;
	upper = root->arc[arc].value + LCB_FACTOR * ucb_var(root, arc);
	if(upper >= lower)	goto dont_cut;
      }
if(2) gprintf("Cut! -- %s.\n", __func__);
      break;
dont_cut:
      continue;
    }
#endif
  } while(!cs_timeup);
  cs_delete_timer(timerid, &sigact);

if(0) { for(int i = 0; i < root->n_arc; i += 1) {
    gprintf("move=%1m v=%.2f%% n=%d\n", root->arc[i].move,
      100 * root->arc[i].value, (int)(root->arc[i].visits + 0.5)); } }

  v = root->arc[bestarc].value;
  if(!gg_never_resign && v < gg_resign_threshold) {
    bestmove = GG_RESIGN_MOVE;
    goto onlyone;
#if 0
  } else if(!gg_never_pass && v >= 1) {
    /* If every move gets 100% of win rate then pass. */
    for(int i = 0; i < root->n_arc; i += 1) {
      if(root->arc[i].value < 1) goto normal;
    }
    goto play_pass;
normal:
    continue;
#endif
  }
  bestmove = root->arc[bestarc].move;
  maxvisit = root->arc[bestarc].visits;

#if USE_OFFSET
  /* TODO Better algorithm */
{
  float delta = 0;
  if(v >= 0.65 && gg_offset < MAX_OFFSET) {
    delta = gg_n_empty_point();
    if(gg_offset < 0) {
      delta = exp10f(1 * ((1.0 / 400) * delta - 1));
    } else {
      delta = exp10f(2 * ((1.0 / 400) * delta - 1));
    }
    gg_offset += delta;
  } else if(v < 0.55 && gg_offset > 0) {
    gg_offset = 0;
  } else if(v < 0.50) {
    delta = gg_n_empty_point();
    delta = -1 * exp10f(1 * ((1.0 / 400) * delta - 1));
    gg_offset += delta;
    if(gg_offset > 0) {
      gg_offset = 0;
    }
  }
  gg_offset = (gg_offset > MAX_OFFSET) ? MAX_OFFSET : gg_offset;
if(2) {
  fprintf(stderr, "Offset=%.1f", gg_offset);
  if(delta != 0) fprintf(stderr, " (%+.1f)", delta);
  fputc('\n', stderr);
}
}
#endif

  gg_back_and_reset_position(root->position_sp);/* Discard simulated worlds. */
if(0) gprintf("-- %s Reset position_sp to %p\n", __func__, root->position_sp);
  goto exit;

#if 0
play_pass:
  bestmove = GG_PASS_MOVE;
  goto onlyone;
#endif

onlyone:
  maxvisit = 0;
  v = 2;	/* TODO Just to distinguish from normal value of v. */
  goto exit;

exit:
#if PROFILING
  gg_print_profile(profile);
#endif
if(2) {
  fprintf(stderr, "Used uct stack = %'u node (%'luMiB)\n", uct_get_n_node(),
	(long unsigned int)uct_get_n_node() * sizeof(ggUctNode) >> 20); }
if(0) gprintf("<-- %s(%C) v=%.1f%% (%1m) visit=%.0f playouts=%u\n",
	__func__, c, 100 * v, bestmove, maxvisit, gg_total_playout);
  return bestmove;
}
