/***************************************************************
 * pack.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "gggo.h"
#include "mc.h"
#include "pack.h"

/* Convert each field of structures to network endian and pack it.
 */
uintptr_t
cs_pack_position(csPackedPosition *dst, const mcPosition *src) {
  const mcRenID max_ren_free = src->max_ren_free;

if(0) fprintf(stderr, "--> %s(%p,%p) max_ren_free=%u\n",
		    __func__, dst, src, max_ren_free);
  /* Scalars */
  for(int i = 0; i < MC_N_COLOR; i += 1) {
    dst->bw_captured[i] = CS_PACK(src->bw_captured[i]);
  }
  dst->movenum = CS_PACK(src->movenum);
  dst->previous_move = CS_PACK(src->previous_move);
  dst->ren_free = CS_PACK(src->ren_free);
  dst->max_ren_free = CS_PACK(src->max_ren_free);
  dst->ko_info.point = CS_PACK(src->ko_info.point);
  dst->ko_info.color = CS_PACK(src->ko_info.color);
  dst->n_pass = CS_PACK(src->n_pass);
  dst->gameend = CS_PACK(src->gameend);
  dst->color = CS_PACK(src->color);
#if USE_LADDER_MODE
  dst->use_ladder_mode = CS_PACK(src->use_ladder_mode);
#endif
  for(int i = 0; i < MC_BOARDSIZE; i += 1) {
    dst->board[i] = CS_PACK(src->board[i]);
    dst->ren_board[i].id = CS_PACK(src->ren_board[i].id);
    dst->ren_board[i].next = CS_PACK(src->ren_board[i].next);
#if USE_EMPTY_LIST
    dst->empty_list[i] = CS_PACK(src->empty_list[i]);
#endif
  }
  for(int i = 0; i < (int)max_ren_free; i += 1) {
    dst->ren[i].next = CS_PACK(src->ren[i].next);
    dst->ren[i].member = CS_PACK(src->ren[i].member);
    dst->ren[i].size = CS_PACK(src->ren[i].size);
    dst->ren[i].liberties = CS_PACK(src->ren[i].liberties);
    dst->ren[i].color = CS_PACK(src->ren[i].color);
    for(int j = 0; j < MC_MAX_LIBS_REN; j += 1) {
      dst->ren[i].liberty[j] = CS_PACK(src->ren[i].liberty[j]);
    }
  }
if(0) fprintf(stderr, "<-- %s(%p,%p) p=%p\n", __func__,
			dst, src, &dst->ren[max_ren_free]);
  return (uintptr_t)&dst->ren[max_ren_free];
}

void
cs_unpack_position(mcPosition *dst, const csPackedPosition *src) {
  const mcRenID max_ren_free = CS_UNPACK(src->max_ren_free);

if(0) fprintf(stderr, "--> %s(%p,%p) max_ren_free=%u\n", __func__,
		dst, src, max_ren_free);
  dst->mark_num = 0;
  dst->mark_ren_num = 0;
  for(int i = 0; i < MC_BOARDSIZE; i += 1) {
    dst->mark[i] = 0;
  }
  for(int i = 0; i < MC_N_COLOR; i += 1) {
    dst->bw_captured[i] = CS_UNPACK(src->bw_captured[i]);
  }
  dst->movenum = CS_UNPACK(src->movenum);
  dst->previous_move = CS_UNPACK(src->previous_move);
  dst->ren_free = CS_UNPACK(src->ren_free);
  dst->max_ren_free = max_ren_free;
  dst->ko_info.point = CS_UNPACK(src->ko_info.point);
  dst->ko_info.color = CS_UNPACK(src->ko_info.color);
  dst->n_pass = CS_UNPACK(src->n_pass);
  dst->gameend = CS_UNPACK(src->gameend);
  dst->color = CS_UNPACK(src->color);
#if USE_LADDER_MODE
  dst->use_ladder_mode = CS_UNPACK(src->use_ladder_mode);
#endif

  for(int i = 0; i < MC_BOARDSIZE; i += 1) {
    dst->board[i] = CS_UNPACK(src->board[i]);
    dst->ren_board[i].id = CS_UNPACK(src->ren_board[i].id);
    dst->ren_board[i].next = CS_UNPACK(src->ren_board[i].next);
#if USE_EMPTY_LIST
    dst->empty_list[i] = CS_UNPACK(src->empty_list[i]);
#endif
  }

  for(int i = 0; i < (int)max_ren_free; i += 1) {
    dst->ren[i].next = CS_UNPACK(src->ren[i].next);
    dst->ren[i].mark = 0;
    dst->ren[i].member = CS_UNPACK(src->ren[i].member);
    dst->ren[i].size = CS_UNPACK(src->ren[i].size);
    dst->ren[i].liberties = CS_UNPACK(src->ren[i].liberties);
    dst->ren[i].color = CS_UNPACK(src->ren[i].color);
    for(int j = 0; j < MC_MAX_LIBS_REN; j += 1) {
      dst->ren[i].liberty[j] = CS_UNPACK(src->ren[i].liberty[j]);
    }
  }
  for(int i = (int)max_ren_free; i < MC_MAX_REN; i += 1) {
    dst->ren[i].next = i + 1;
    dst->ren[i].mark = 0;
#if SAFETY
    dst->ren[i].member = 0;
    dst->ren[i].size = 0;
    dst->ren[i].liberties = 0;
    dst->ren[i].color = MC_EMPTY;
    for(int j = 0; j < MC_MAX_LIBS_REN; j += 1) {
      dst->ren[i].liberty[j] = 0;
    }
#endif
  }
  dst->ren[MC_MAX_REN - 1].next = 0;
if(0) fprintf(stderr, "<-- %s(%p,%p)\n", __func__, dst, src);
  return;
}

#if USE_RAVE
uintptr_t
cs_pack_rave(csRave *dst, const mcRave *src) {
  const int n = (int)src->n_rave;
  mcMove *p;
  mcColor *q;

  dst->n_rave = CS_PACK((int16_t)n);
  p = (mcMove *)&dst->space[0];
  for(int i = 0; i < n; i += 1) {
    *p++ = CS_PACK(src->move[i]);
  }
  q = (mcColor *)p;
  for(int i = 0; i < n; i += 1) {
    *q++ = CS_PACK(src->color[i]);
  }
  return (uintptr_t)q;
}

void
cs_unpack_rave(mcRave *dst, const csRave *src) {
  const int n = CS_UNPACK(src->n_rave);
  const mcMove *p;
  const mcColor *q;

  dst->n_rave = (int16_t)n;
  p = (const mcMove *)&src->space[0];
  for(int i = 0; i < n; i += 1) {
    dst->move[i] = CS_UNPACK(*p++);
  }
  q = (const mcColor *)p;
  for(int i = 0; i < n; i += 1) {
    dst->color[i] = CS_UNPACK(*q++);
  }
  return;
}
#endif	/* USE_RAVE */

void
dump_mc_position(mcPosition *p, mcMove first_move, FILE *s) {
  int i;
  gfprintf(s, "DumpPosition:"
#if 0
      " gen/po=%6u/%6u board_size=%u po/node=%6u/%6u optarc=%d"
      " t_send/recv=%6u/%6u"
#endif
      " firstmove=%C(%1m) movenum=%u n_pass=%u"
      " previous_move=%1m ko_info=%1m(%C) captured=%u/%u mark_num=%u"
      " mark_ren_num=%u ren_free=%u max_ren_free=%u gameend=%b",
#if 0
      p->generation, p->playout, p->board_size, p->node, p->optarc,
      p->t_send, p->t_recv,
#endif
      p->color, first_move, p->movenum, p->n_pass,
      p->previous_move, p->ko_info.point, p->ko_info.color,
      p->bw_captured[0], p->bw_captured[1],
      p->mark_num, p->mark_ren_num, p->ren_free, p->max_ren_free,
      p->gameend);
#if USE_LADDER_MODE
  gfprintf(s, " ladder_mode=%b", p->use_ladder_mode);
#endif
  fputs("\nRen_board:", s);
  for(i = 0; i < MC_BOARDSIZE; i += 1) {
    fprintf(s, " %u/%u", p->ren_board[i].id, p->ren_board[i].next);
  }
  fputs("\nBoard:", s);
  for(i = 0; i < MC_BOARDSIZE; i += 1) {
    gfprintf(s, " %C", p->board[i]);
  }
  putc('\n', s);
#if USE_EMPTY_LIST
  fputs("Empty_list:", s);
#if 0
  for(i = 0; i < MC_BOARDSIZE; i += 1) {
    gfprintf(s, " %1m", p->empty_list[i]);
  }
#else
  for(i = 0; p->empty_list[i] != 0; i = p->empty_list[i]) {
    gfprintf(s, " %1m", p->empty_list[i]);
  }
#endif
  putc('\n', s);
#endif
#if USE_ADJACENT_REN
  /* TODO */
#endif
#if 0
  for(i = 0; i < MC_MAX_REN; i += 1) {
    gfprintf(s, "Ren[%2d]: "
	  "next=%u member=%1m size=%u liberties=%u mark=%u color=%C",
	  i, p->ren[i].next, p->ren[i].member, p->ren[i].size,
	  p->ren[i].liberties, p->ren[i].mark, p->ren[i].color);
    fputs(" liberty=", s);
    for(int j = 0; j < MC_MAX_LIBS_REN; j += 1) {
      gfprintf(s, " %1m", p->ren[i].liberty[j]);
    }
    putc('\n', s);
  }
#else
  for(i = 0; (i = p->ren[i].next) != 0; ) {
    gfprintf(s, "Ren[%2d]: "
	  "next=%u member=%1m size=%u liberties=%u mark=%u color=%C",
	  i, p->ren[i].next, p->ren[i].member, p->ren[i].size,
	  p->ren[i].liberties, p->ren[i].mark, p->ren[i].color);
    fputs(" liberty=", s);
    for(int j = 0; j < MC_MAX_LIBS_REN; j += 1) {
      gfprintf(s, " %1m", p->ren[i].liberty[j]);
    }
    putc('\n', s);
  }
#endif
#if 0
#if USE_RAVE
  fprintf(s, "Rave: n_rave=%d", p->rave.n_rave);
  for(int i = 0; i < p->rave.n_rave; i += 1) {
    gfprintf(s, " %1m(%C)", p->rave.move[i], p->rave.color[i]);
  }
  putc('\n', s);
#endif
#endif
  fflush(s);
  return;
}

#if USE_RAVE
void
dump_mc_rave(mcRave *p, FILE *s) {
  fprintf(s, "DumpMCRave: n=%d", p->n_rave);
  for(int i = 0; i < p->n_rave; i += 1) {
    gfprintf(s, " %1m(%C)", p->move[i], p->color[i]);
  }
  putc('\n', s);
  fflush(s);
  return;
}

void
dump_cs_rave(csRave *p, FILE *s) {
  const int n = (int)CS_UNPACK(p->n_rave);
  mcColor *q;

  q = (mcColor *)((uintptr_t)p + sizeof(uint16_t) + n * sizeof(mcMove));
  fprintf(s, "DumpCSRave: n=%d", n);
  for(int i = 0; i < n; i += 1) {
    gfprintf(s, " %1m(%C)", CS_UNPACK(p->move[i]), CS_UNPACK(q[i]));
  }
  putc('\n', s);
#if 0
  q = (mcColor *)((uintptr_t)p + sizeof(uint16_t));
  fprintf(s, "DumpBinary: n=%d", n);
  for(unsigned i = 0; i < sizeof(csRave) - sizeof(p->n_rave); i += 1) {
    fprintf(s, " 0x%2x", q[i]);
  }
  putc('\n', s);
#endif
  fflush(s);
  return;
}
#endif
