/***************************************************************
 * play.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <string.h>
#include "gggo.h"
#include "mc.h"
#include "queue.h"
#include "play.h"

/************************
 *	Interface	*
 ************************/
extern void	gg_push_clock_button(ggColor color, ggMove move);

#if USE_HASH

#define GG_N_TRANS	(64 * 1024)
#define GG_H2X(hash)	(((hash) + (hash >> 32)) & (GG_N_TRANS - 1))

void	gg_create_hash(void);
void	gg_init_hash(void);
void	gg_report_hash(FILE *stream);
void	gg_init_trans(void);
bool	gg_is_duplicated(ggMove move, ggColor color);

ggHash		gg_hash;

static ggPosition *gg_trans[GG_N_TRANS];

#endif

/********************************************************
 *		Handle multiple worlds			*
 ********************************************************/
#if USE_DYNAMIC_ALLOCATE
static ggPositionAve	*gg_position;
ggPosition	*gg_position_sp, *gg_position_limit;

char *	gg_init_position(uint32_t n);
char *
gg_init_position(uint32_t n) {
  int rc;
if(0) gprintf("--> %s(%u)\n", __func__, n);
  rc = posix_memalign((void **)&gg_position, sysconf(_SC_PAGESIZE),
			  (size_t)n * sizeof(ggPositionAve));
  gg_position_sp = (ggPosition *)&gg_position[0];
  gg_position_limit = (ggPosition *)&gg_position[n];
if(2) gprintf("Allocate position stack: size=%ldMiB top=%p limit=%p\n",
	    (n * sizeof(ggPositionAve) + (1L << 20) - 1) >> 20,
	    gg_position, gg_position_limit);
if(0) gprintf("<-- %s(%u) rc=%d\n", __func__, n, rc);
  return rc == 0 ? NULL : strerror(rc);
}
#else
static ggPositionAve	gg_position[GG_MAX_POSITION];
ggPosition	*gg_position_sp = (ggPosition *)&gg_position[0];
#endif

/* Copy global variables back to top of position stack.
 */

ggPosition *
gg_push_position(void) {
extern uint32_t	uct_get_n_node(void);
  /* Increment sp (variable sized).
   */
  ggPosition *sp = gg_position_sp++;
if(0) gprintf("--> %s()=%d\n", __func__);
  gg_position_sp = (ggPosition *)((ggRen *)gg_position_sp + gg_max_ren_free);
#if USE_DYNAMIC_ALLOCATE
  if(gg_position_sp >= gg_position_limit)
#else
  if(gg_position_sp >= (ggPosition *)&gg_position[GG_MAX_POSITION])
#endif
      GG_FATAL("Position stack overflow sp=%p uct=%u\n",
			gg_position_sp, uct_get_n_node());
  gg_position_sp->previous_sp = sp;
  sp->next_sp = gg_position_sp;
#if USE_HASH
  sp->hash = gg_hash;
#endif
  sp->mark_num = gg_mark_num;
  sp->mark_ren_num = gg_mark_ren_num;
  sp->bw_captured[GG_C2X(GG_WHITE)] = gg_bw_captured[GG_C2X(GG_WHITE)];
  sp->bw_captured[GG_C2X(GG_BLACK)] = gg_bw_captured[GG_C2X(GG_BLACK)];
  sp->movenum = gg_movenum;
  sp->board_ko_info = gg_board_ko_info;
  sp->previous_move = gg_previous_move;
  sp->previous_move2 = gg_previous_move2;
  sp->ren_free = gg_ren_free;
  sp->max_ren_free = gg_max_ren_free;
  sp->n_pass = gg_n_pass;
  sp->gameend = gg_gameend;
#if USE_EMPTY_LIST
  sp->empty_list0 = gg_empty_list[0];
#endif
  memcpy(&sp->board[0], &gg_board[0], sizeof gg_board);
#if USE_EMPTY_LIST
  memcpy(&sp->empty_list[0], &gg_empty_list[0], sizeof gg_empty_list);
#endif
  for(ggRenID r = 0; r < gg_max_ren_free; r += 1) {
    sp->ren[r] = gg_ren[r];
  }
#if USE_ADJACENT_REN
  for(ggRenID r = 0; r < gg_max_ren_free; r += 1) {
    for(ggRenID s = 0; s < gg_max_ren_free; s += 1) {
      gg_position_sp->adjacent_ren[r][s] = sp->adjacent_ren[r][s];
    }
  }
#endif
if(0) gprintf("<-- %s sp=%p\n", __func__, sp);
  return sp;
}

/* Switch back to the world pointed by sp.
 * Create new frame if necessary.
 */
void
gg_backto_position(ggPosition *sp) {
if(0) gprintf("--> %s(%p) position_sp=%p\n", __func__, gg_position_sp);
#if USE_HASH
  gg_hash = sp->hash;
#endif
  gg_mark_num = sp->mark_num;
  gg_mark_ren_num = sp->mark_ren_num;
  gg_bw_captured[GG_C2X(GG_WHITE)] = sp->bw_captured[GG_C2X(GG_WHITE)];
  gg_bw_captured[GG_C2X(GG_BLACK)] = sp->bw_captured[GG_C2X(GG_BLACK)];
  gg_board_ko_info = sp->board_ko_info;
  gg_movenum = sp->movenum;
  gg_ren_free = sp->ren_free;
  gg_max_ren_free = sp->max_ren_free;
  gg_previous_move = sp->previous_move;
  gg_previous_move2 = sp->previous_move2;
  gg_n_pass = sp->n_pass;
  gg_gameend = sp->gameend;

  for(ggMove p = 0; p < GG_BOARDSIZE; p += 1) {
    gg_board[p].mark = 0;
#if SAFETY
    gg_board[p].ren.id = 0;
    gg_board[p].ren.next = 0;
    gg_board[p].color = GG_EXTER;
#if USE_EMPTY_LIST
    gg_empty_list[p] = 0;
#endif
#endif
  }
  memcpy(&gg_board[0], &sp->board[0], sizeof gg_board);
#if USE_EMPTY_LIST
  memcpy(&gg_empty_list[0], &sp->empty_list[0], sizeof gg_empty_list);
  gg_empty_list[0] = sp->empty_list0;
#endif
  for(ggRenID i = 0; i < gg_max_ren_free; i += 1) {
    gg_ren[i] = sp->ren[i];
  }
  for(ggRenID i = gg_max_ren_free; i < GG_MAX_REN; i += 1) {
    gg_ren[i].next = i + 1;
    gg_ren[i].mark = 0;
#if SAFETY
    gg_ren[i].member = 0;
    gg_ren[i].size = 0;
    gg_ren[i].liberties = 0;
    gg_ren[i].color = GG_EMPTY;
    for(uint16_t j = 0; j < GG_MAX_LIBS_REN; j += 1) {
      gg_ren[i].liberty[j] = 0;
    }
#endif
  }
  gg_ren[GG_MAX_REN - 1].next = 0;
if(0) gprintf("<-- %s max_ren_free=%d position_sp=%p\n", __func__,
	  gg_max_ren_free, gg_position_sp);
  return;
}

void
gg_back_and_reset_position(ggPosition *sp) {
  gg_backto_position(sp);
  gg_position_sp = sp;
  return;
}

/* Undo
 */
static void
gg_pop_position(void) {
if(0) gprintf("--> %s() sp=%p\n", __func__, gg_position_sp);
  if(gg_position_sp->previous_sp == NULL) {
    GG_FATAL("Position stack underflow\n");
  }
  gg_backto_position(gg_position_sp->previous_sp);
  gg_position_sp = gg_position_sp->previous_sp;
if(0) gprintf("<-- %s sp=%p\n", __func__, gg_position_sp);
  return;
}

void
gg_init_stack(void) {
if(0) gprintf("--> %s()\n", __func__);
  gg_position_sp = (ggPosition *)&gg_position[0];
  gg_position_sp->previous_sp = NULL;
if(0) gprintf("<-- %s sp=%p\n", __func__, gg_position_sp);
  return;
}

bool
gg_undo_move(int n) {
  ggPosition *sp = gg_position_sp;
  while(--n > 0) {
    sp = sp->previous_sp;
    if(sp == NULL)	return false;
  }
  gg_position_sp = sp;
  gg_pop_position();
  return true;
}

/****************************************
 *    Macros for including SFMT.h	*
 ****************************************/
#define genrand_real2	gg_random
#define gen_rand32	gg_rand32
#define gen_rand64	gg_rand64

#define get_idstring	gg_get_idstring
#define get_min_array_size32	gg_get_min_array_size32
#define get_min_array_size64	gg_get_min_array_size64
#define fill_array32	gg_fill_array32
#define fill_array64	gg_fill_array64
#define init_gen_rand	gg_init_gen_rand
#define init_by_array	gg_init_by_array

#include "SFMT.h"
#include "SFMT.c"

#if USE_HASH
/********************************
 *	Zobrist hash		*
 ********************************/
ggHash	gg_zobrist[GG_BOARDSIZE][GG_N_COLOR];
static int hash_total, hash_collision, hash_hit;

void
gg_create_hash(void) {
  ggColor c;
  ggMove m;

if(0) gprintf("--> %s()\n", __func__);
  GG_FOR_ALL_POINT(m) {
    GG_FOR_ALL_COLOR(c) {
      gg_zobrist[m][GG_C2X(c)] = gg_rand64();
    }
  }
if(0) {
  GG_FOR_ALL_POINT(m) {
    GG_FOR_ALL_COLOR(c) {
      printf(" %#lx", gg_zobrist[m][GG_C2X(c)]);
    }
  }
  putchar('\n');
}
if(0) gprintf("<-- %s()\n", __func__);
  return;
}

void
gg_init_hash(void) {
  gg_hash = 0;
  hash_total = hash_collision = hash_hit = 0;
  return;
}

void
gg_init_trans(void) {
  for(int i = 0; i < GG_N_TRANS; i += 1) {
    gg_trans[i] = NULL;
  }
  return;
}

bool
gg_is_duplicated(ggMove move, ggColor color) {
  ggHash key;
  ggPosition *sp;
  bool rc;

if(0) gprintf("--> %s(%1m,%C)\n", __func__, move, color);
  assert(gg_is_not_ob(move));
  assert(gg_is_onboard(move));
  hash_total += 1;
  rc = false;
  key = gg_hash ^ gg_zobrist[move][GG_C2X(color)];
  sp = gg_trans[GG_H2X(key)];
  if(sp != NULL && sp->hash == key) {
#if 0
    ggMove p;
if(0) gprintf("Same key sp=%p at %1m(%C) curr_sp=%p -- %s.\n",
	      sp, move, color, gg_position_sp, __func__);
    /* Play and compare boards. */
    gg_push_position();
    gg_do_play_move(move, color);
    GG_FOR_ALL_POINT(p) {
      if(gg_get_color(p) != sp->board[p].color) {
	hash_collision += 1;
	goto undo;
      }
    }
#endif
    rc = true;
    hash_hit += 1;
if(0) gprintf("Duplicated sp=%p at %1m(%C) -- %s.\n", move, color, __func__);
#if 0
undo:
    gg_pop_position();
#endif
if(0) gprintf("Undo sp=%p -- %s.\n", gg_position_sp, __func__);
  }
if(0) gprintf("<-- %s %B\n", __func__, rc);
  return rc;
}

void
gg_report_hash(FILE *stream) {
  float all, hit, collision;
  all = (float)hash_total;
  hit = (float)hash_hit / all;
  collision = (float)hash_collision / all;
  fprintf(stream, "Hash statistics: hit=%d(%f%%), collision=%d(%f%%), all=%d\n",
      hash_hit, hit * 100, hash_collision, collision * 100, hash_total);
  return;
}
#endif	/* USE_HASH */

/********************************
 *	Play the move.		*
 ********************************/
void
gg_do_play_move(ggMove move, ggColor color) {

if(0) gprintf("--> %s(%1m,%C)\n", __func__, move, color);

  assert(!gg_is_illegal(move, color));
#if SAFETY
  gg_gameend = false;
#endif
  if(move == GG_PASS_MOVE) {
    if(++gg_n_pass >= 2) {
      gg_gameend = true;
    }
  } else {
    gg_n_pass = 0;
    gg_play_and_analyze(move, color);
  }
  gg_previous_move2 = gg_previous_move;
  gg_previous_move = move;
  gg_movenum += 1;
if(0) gprintf("<-- %s\n", __func__);
  return;
}

void
gg_play_move(ggMove move, ggColor color) {
if(0) gprintf("--> %s(%1m,%C)\n", __func__, move, color);
  gg_push_clock_button(color, move);
  gg_push_position();
  gg_do_play_move(move, color);
#if USE_HASH
if(1) gprintf("Play %C at %1m, hash=%#x, num=%d.\n",
		  color, move, gg_hash, gg_movenum);
#endif
if(0) gprintf("<-- %s\n", __func__);
  return;
}

/* Assume all empty points are single eyes.
 * If not, result may contain some error.
 * FIXME Manage eye spaces and seki properly.
 */
int
gg_count_final_score(ggColor color, int *n_dame0) {
  int score;
  ggMove p;
  ggColor c;

if(0) gprintf("--> %s(%C,%p)\n", __func__, color, n_dame0);
  /* TODO Use SIMD insn. */
  score = 0;
  if(n_dame0 != NULL)	*n_dame0 = 0;
  GG_FOR_ALL_POINT(p) {
    c = gg_get_color(p);
    switch(c) {
    case GG_EMPTY:
      c = gg_get_color(GG_NORTH(p));
      if(c == GG_BLACK || c == GG_WHITE)	goto count;
      c = gg_get_color(GG_EAST(p));
      if(c == GG_BLACK || c == GG_WHITE)	goto count;
      c = gg_get_color(GG_SOUTH(p));
      if(c == GG_BLACK || c == GG_WHITE)	goto count;
      c = gg_get_color(GG_WEST(p));
      if(c == GG_BLACK || c == GG_WHITE)	goto count;
      if(n_dame0 != NULL)	*n_dame0 += 1;
      break;
    case GG_BLACK: case GG_WHITE:
count:
      score += (c == color) ? 1 : -1;
      break;
    case GG_EXTER:
      break;
    default:
      GG_FATAL("Impossible color=%d at %1m\n", gg_get_color(p), p);
      break;
    }
  }
if(0) gprintf("<-- %s score=%+3d,dame=%d\n", __func__, score,
	  n_dame0 == NULL ? -1 : *n_dame0);
  return score;
}

/* FIXME */
ggScoreValue
gg_finish_and_score(ggRandomSeed seed) {
  ggScoreValue final_score;
  UNUSED(seed);
if(0) gprintf("--> %s(%u)\n", __func__, seed);
  /* TODO Play best moves until end. */
  final_score = gg_count_final_score(GG_WHITE, NULL) + gg_komi;
if(0) gprintf("<-- %s score=%+3d\n", __func__, final_score);
  return final_score;
}

bool
gg_is_self_atari(ggMove m, ggColor color) {
  int libs;
  bool rc;
  union {
    uint8_t	b[4];
    uint32_t	i;
  } adjacent;
  ggMove q;
  ggRenID s;

if(0) gprintf("<-> %s(%1m, %C)", __func__, m, color);
  assert(gg_is_not_ob(m));
  assert(gg_is_empty(m));
  rc = false;
#if 0
  for(int i = 0; i < 4; i += 1) adjacent.b[i] = 0;
#else
  adjacent.i = 0;
#endif
  GG_FOR_4_ADJACENT_POINT(q, m) {
    adjacent.b[gg_get_color(q)]	+= 1;
  }
  libs = adjacent.b[GG_EMPTY];
  if(libs >= 2)		goto fail;

  /* Can capture? */
  if(adjacent.b[gg_opposit_color(color)] > 0) {
    GG_FOR_4_ADJACENT_POINT(q, m) {
      if(!gg_is_opponent(q, color))	continue;
      s = gg_id_ren(q);
      assert(s != 0);
      if(gg_libs_ren(s) <= 1)	goto fail;
    }
  }

  /* No. Can connect friends?
   */
  if(adjacent.b[color] > 0) {
#if 1
    gg_reset_renmark();
    gg_reset_mark();
    gg_mark(m);
    GG_FOR_4_ADJACENT_POINT(q, m) {
      if(!gg_is_friend(q, color))		continue;
      s = gg_id_ren(q);
      assert(s != 0);
      if(gg_is_marked_ren(s))	continue;
      for(int i = 0; i < gg_libs_ren(s); i += 1) {
      	if(i < GG_MAX_LIBS_REN) {
	  ggMove x = gg_liberty_ren(s)[i];
	  assert(gg_is_empty(x));
	  if(gg_is_marked(x))	continue;
	  gg_mark(x);
	}
	libs += 1;
	if(libs >= 2)		goto fail;
      }
      gg_mark_ren(s);
    }
#else
{
    ggMove liberty[4];
    int n = 0;
    gg_reset_renmark();
    liberty[n++] = m;
    GG_FOR_4_ADJACENT_POINT(q, m) {
      if(!gg_is_friend(q, color))		continue;
      s = gg_id_ren(q);
      assert(s != 0);
      if(gg_is_marked_ren(s))	continue;
      for(int i = 0; i < gg_libs_ren(s); i += 1) {
      	if(i < GG_MAX_LIBS_REN) {
	  ggMove x = gg_liberty_ren(s)[i];
	  assert(gg_is_empty(x));
	  for(int j = 0; j < n; j += 1) {
	    if(x == liberty[j])	goto next_liberty;
	  }
	}
	libs += 1;
	if(libs >= 2)		goto fail;
	assert(n < 4);
	liberty[n++] = x;
next_liberty: ;
      }
      gg_mark_ren(s);
    }
}
#endif
  }
  rc = true;
fail:
if(0) gprintf(" %b\n", rc);
  return rc;
}

/* Setup mc_position from p->position.
 */
void
gg_setup_position(mcPosition *position, ggPosition *sp, ggColor color) {
  ggRenID ren_free, max_ren_free;

if(0) gprintf("--> %s(%p,%p,%C)", __func__, position, sp, color);
  position->bw_captured[MC_C2X(MC_WHITE)] = sp->bw_captured[GG_C2X(GG_WHITE)];
  position->bw_captured[MC_C2X(MC_BLACK)] = sp->bw_captured[GG_C2X(GG_BLACK)];
  position->movenum = sp->movenum;
  position->ko_info = sp->board_ko_info;
  position->color = color;
  position->n_pass = sp->n_pass;
  position->previous_move = sp->previous_move;
  ren_free = position->ren_free = sp->ren_free;
  max_ren_free = position->max_ren_free = sp->max_ren_free;
  position->gameend = sp->gameend;
#if USE_LADDER_MODE
  position->use_ladder_mode = gg_use_ladder_mode;
#endif

  position->mark_num = 0;
  for(ggPoint p = 0; p < MC_BOARDSIZE; p += 1) {
    position->mark[p] = 0;
#if SAFETY
    position->board[p] = MC_EXTER;
    position->ren_board[p].id = 0;
    position->ren_board[p].next = 0;
#if USE_EMPTY_LIST
    position->empty_list[p] = 0;
#endif
#endif
  }
  for(int x = 0; x < gg_board_size; x += 1) {
    for(int y = 0; y < gg_board_size; y += 1) {
      position->board[MC_POS(x, y)] = sp->board[GG_POS(x, y)].color;
      position->ren_board[MC_POS(x, y)] = sp->board[GG_POS(x, y)].ren;
#if USE_EMPTY_LIST
      position->empty_list[MC_POS(x, y)] = sp->empty_list[GG_POS(x, y)];
#endif
    }
  }
#if USE_EMPTY_LIST
  position->empty_list[0] = sp->empty_list0;
#endif

  position->mark_ren_num = 0;
  for(int i = 0; i < max_ren_free; i += 1) {
    position->ren[i] = sp->ren[i];
    position->ren[i].mark = 0;
  }
  for(int i = max_ren_free; i < MC_MAX_REN; i += 1) {
    position->ren[i].next = i + 1;
    position->ren[i].mark = 0;
#if SAFETY
    position->ren[i].color = MC_EMPTY;
    position->ren[i].member = 0;
    position->ren[i].size = 0;
    position->ren[i].liberties = 0;
    for(int j = 0; j < MC_MAX_LIBS_REN; j += 1) {
      position->ren[i].liberty[j] = 0;
    }
#endif
  }
  position->ren[GG_MAX_REN - 1].next = 0;

if(0) gprintf("<-- %s(%p,%p,%C)", __func__, position, sp, color);
  return;
}
