/***************************************************************
 * timebase.h
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef _TIMEBASE_H_
#define _TIMEBASE_H_

#include <stdint.h>

#if defined(__SPU__)

#include <spu_mfcio.h>
#define TIMEBASE		79800000	/* For PS3 */
#define SPU_DECREMENTER_INITIAL_VALUE	0x7fffffffu
#define START_CLOCK()	\
  (spu_writech(SPU_WrDec, SPU_DECREMENTER_INITIAL_VALUE),	\
   (clock_t)spu_readch(SPU_RdDec))
#define DELTA_CLOCK(t)		((clock_t)((t) - spu_readch(SPU_RdDec)))
extern inline uint32_t	profile_start(void);
extern inline uint32_t
profile_start(void) {
  spu_writech(SPU_WrDec, SPU_DECREMENTER_INITIAL_VALUE);
  return spu_readch(SPU_RdDec);
}
extern inline uint32_t	profile_delta(uint32_t ts);
extern inline uint32_t
profile_delta(uint32_t ts) { return ts - spu_readch(SPU_RdDec); }

#elif defined(__PPU__)

#define TIMEBASE		79800000	/* For PS3 */
#define START_CLOCK()		get_timebase_lower32()
#define DELTA_CLOCK(t)		(get_timebase_lower32() - (t))
extern inline uint32_t get_timebase_lower32(void);
extern inline uint32_t
get_timebase_lower32(void) {
  register uint32_t timebase_lower32;
  __asm__ volatile("mftb %0" : "=r"(timebase_lower32));
  return timebase_lower32;
}
#if 0
extern inline uint32_t get_timebase_upper32(void);
extern inline uint32_t
get_timebase_upper32(void) {
  register uint32_t timebase_upper32;
  __asm__ volatile("mftbu %0" : "=r"(timebase_upper32));
  return timebase_upper32;
}
extern inline uint64_t get_timebase64(void);
extern inline uint64_t
get_timebase64(void) {
  register uint32_t timebase_upper32, timebase_lower32;
  do {
    timebase_upper32 = get_timebase_upper32();
    timebase_lower32 = get_timebase_lower32();
  } while(get_timebase_upper32() != timebase_upper32);
  return (uint64_t)timebase_upper32 << 32 | (uint64_t)timebase_lower32;
}
#endif
extern inline uint32_t profile_start(void);
extern inline uint32_t
profile_start(void)		{ return get_timebase_lower32(); }
extern inline uint32_t profile_delta(uint32_t ts);
extern inline uint32_t
profile_delta(uint32_t ts)	{ return get_timebase_lower32() - ts; }

#else	/* Other CPUs */

#define TIMEBASE		CLOCKS_PER_SEC
#define START_CLOCK()		(clock())
#define DELTA_CLOCK(t)		(clock() - (t))

#endif

#define MILLISECOND(t)		(1e3 * (t) / (float)TIMEBASE)
#define SECOND(t)		((float)(t) / (float)TIMEBASE)

#endif /* _TIMEBASE_H_ */
