/***************************************************************
 * uct.h
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef __UCT_H__
#define __UCT_H__

#include "gggo.h"

/********************************
 *	Node of UCT Tree	*
 ********************************/
typedef double	ggSigma;

typedef struct uct_node {
  ggPosition	*position_sp;
  struct uct_node	*p_node;	/* *Parent node */
  float		visits;		/* total visits */
#if USE_RAVE
  float		m_rave;		/* total visits_rave */
#endif
  ggScore	score;
  ggArc		p_arc;		/* Parent move */
  ggArc		n_arc;		/* Number of legal moves */
#if USE_ARC_CACHE
  ggSArc	optarc;		/* Cache of optimal move. */
#endif
  ggColor	color;		/* Color (player) of this node */
  struct uct_arc {
    struct uct_node	*node;
    ggSigma	value;
#if USE_UCB1_TUNED
    ggSigma	value2;		/* value^2 */
#endif
#if USE_RAVE
    ggSigma	q_rave;		/* value_rave */
#if USE_UCB1_TUNED
    ggSigma	q_rave_sqr;	/* value_rave^2 */
#endif
#endif
    float	visits;
    float	ucb;
#if USE_RAVE
    float	m_rave;		/* visits_rave */
#endif
#if USE_INDIVIDUAL_FACTOR
    float	factor;
#endif
    ggMove	move;
  } arc[GG_MAX_MOVE];
} ggUctNode;

typedef struct uct_arc	ggUctArc;

#define UCT_NTH(p)	((p) - gg_uct_node)

#if USE_DYNAMIC_ALLOCATE
extern ggUctNode	*gg_uct_node;
#else
extern ggUctNode	gg_uct_node[GG_MAX_UCT_NODE];
#endif

extern ggMove	gg_genjoseki(int movenum);
extern uint32_t	uct_get_n_node(void);
extern void	uct_init_tree(void);
extern bool	uct_is_wrong_node(uint32_t node);
extern ggSArc	uct_select_optarc(ggUctNode *p);
extern ggUctNode *uct_create_node(ggColor color, ggPosition *sp,
			ggUctNode *p_node, ggArc p_arc);
extern void	uct_update_tree(ggUctNode *node, ggArc arc, float value);
#if USE_RAVE
extern void	rave_update_tree(ggUctNode *leaf, ggArc optarc,
		      float value, ggColor color, const mcRave *rave);
#endif

#endif	/* __UCT_H__ */
