/***************************************************************
 * update_ren.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/****************************************
 *	Common Functions for Ren	*
 ****************************************/

#define ADD_DIRTY_IF(s)	do {	\
  assert((s) != 0);		\
  if(!is_marked_ren(s))	{ *dirty_ren0++ = (s); mark_ren(s); }	\
  } while(0)

#define CAPTURE(opp) do {		\
    assert(opp != 0);			\
    assert(libs_ren(opp) > 0);		\
    if(libs_ren(opp) <= 1) {		\
      captured_pos = member_ren(opp);	\
      capture += remove_ren(opp);	\
    } else {				\
      ADD_DIRTY_IF(opp);		\
    }					\
  } while(0)

#if 1
#define CAPTURE3(opp1, opp2, opp3) do {	\
    opp1 = id_ren(opp1);	\
    opp2 = id_ren(opp2);	\
    opp3 = id_ren(opp3);	\
    if(opp1 == opp2) {		\
      if(opp1 == opp3) {	\
	CAPTURE(opp1);		\
      } else {			\
	CAPTURE(opp1);		\
	CAPTURE(opp3);		\
      }				\
    } else if((opp1 == opp3) || (opp2 == opp3)) {	\
      CAPTURE(opp1);		\
      CAPTURE(opp2);		\
    } else {			\
      CAPTURE(opp1);		\
      CAPTURE(opp2);		\
      CAPTURE(opp3);		\
    }				\
  } while(0)
#else
#define CAPTURE3(opp1, opp2, opp3) do {		\
    assert(get_color(opp1) == opposit_color(c));	\
    opp1 = id_ren(opp1);			\
    CAPTURE(opp1);				\
    if(get_color(opp2) == opposit_color(c)) {	\
      opp1 = id_ren(opp2);			\
      if(!is_marked_ren(opp2))	CAPTURE(opp2);	\
    }						\
    if(get_color(opp3) == opposit_color(c)) {	\
      opp1 = id_ren(opp3);			\
      if(!is_marked_ren(opp3))	CAPTURE(opp3);	\
    }						\
  } while(0)
#endif

#if USE_EMPTY_LIST
static void
remove_from_empty_list(const Move m) {
  Move p, q;
if(0) gprintf("--> %s(%1m)\n", __func__, m);
  for(p = 0; (q = empty_list[p]) != 0 && q != m; p = q);
  assert(q == m);
  empty_list[p] = empty_list[m];
if(0) gprintf("<-- %s p=%d q=%d\n", __func__, p, q);
  return;
}
#endif

static INLINE Move
last_stone(const RenID s) {
  Move p, q;
  assert(s != 0);
  assert(libs_ren(s) > 0);
  for(p = member_ren(s); (q = next_member_ren(p)) != 0; p = q);
  return p;
}

static INLINE RenID
create_ren(const Move p, const Color c, const int libs) {
  const RenID new = ren_free;

if(0) gprintf("--> %s(%1m,%C)\n", __func__, p, c);
  if(new == 0)	 FATAL("Free-list exhausted free=%d\n", ren_free);
  ren_free = next_ren(new);
#ifdef max_ren_free
  max_ren_free = (ren_free > max_ren_free) ? ren_free : max_ren_free;
#endif
  /* Put new item onto active ren list. */
  next_ren(new) = next_ren(0);
  next_ren(0) = new;
  color_ren(new) = c;
  size_ren(new) = 1;
  libs_ren(new) = libs;
  member_ren(new) = p;
  next_member_ren(p) = 0;
#if USE_ADJACENT_REN
{ Move q;
  FOR_4_ADJACENT_POINT(q, p) {
    RenID s;
    if(!is_b_or_w(q)) continue;
    s = id_ren(q);
    assert(s != 0);
    adjacent_ren[s][new] = true;
    adjacent_ren[new][s] = true;
  }
}
#endif
if(0)
#ifdef max_ren_free
  gprintf("<-- %s new=%d free=%d max_free=%d\n", __func__,
	new, ren_free, max_ren_free);
#else
  gprintf("<-- %s new=%d free=%d\n", __func__, new, ren_free);
#endif
  return new;
}

static INLINE int
collect_ren(const RenID s) {
  int size;
  RenID t, u;

if(0) gprintf("--> %s(%d)\n", __func__, s);
  assert(s != 0);
  assert(libs_ren(s) > 0);
  size = size_ren(s);
  /* Remove s from all-ren-list which linked from ren[0].next.
   */
  for(t = 0; (u = next_ren(t)) != 0 && u != s; t = u);
  assert(u == s);
  next_ren(t) = next_ren(s);
  color_ren(s) = EMPTY;
#if SAFETY	/* This is not necessary but safety. */
  size_ren(s) = 0;
  libs_ren(s) = 0;
  member_ren(s) = 0;
  for(int i = 0; i < MAX_LIBS_REN; i += 1) {
    (liberty_ren(s))[i] = 0;
  }
#endif
/* Atomic */
  next_ren(s) = ren_free;
  ren_free = s;
/* operation */
if(0) gprintf("<-- %s size=%d\n", __func__, size);
  return size;
}

static INLINE int
remove_ren(const RenID s) {
  Move p, q;
  RenID t;
  int size;
#if USE_EMPTY_LIST
  Move tail;
#endif

if(0) gprintf("--> %s(%d)\n", __func__, s);
  /* Add adjacent rens to dirty list to count
   * their liberties later.
   */
  assert(s != 0);
  assert(libs_ren(s) > 0);
#if USE_EMPTY_LIST
  p = 0;
  do {
    tail = p;
    p = empty_list[p];
  } while(p != 0);
#endif
  FOR_ALL_POINT_IN_REN(p, s) {
    clear_color(p);
#if USE_EMPTY_LIST
#if 0	/* Insert to top of list */
    empty_list[p] = empty_list[0];
    empty_list[0] = p;
#else
    empty_list[tail] = p;
    empty_list[p] = 0;
    tail = p;
#endif
#endif
#if SAFETY
    id_ren(p) = 0;
#endif
    FOR_4_ADJACENT_POINT(q, p) {
      if(is_b_or_w(q)) {
	t = id_ren(q);
	assert(t != 0);
	ADD_DIRTY_IF(t);
      }
    }
  }
#if USE_ADJACENT_REN
  for(t = 0; t < max_ren_free; t += 1) {
    adjacent_ren[s][t] = false;
    adjacent_ren[t][s] = false;
  }
#endif
  /* Then collect entry. */
  size = collect_ren(s);
if(0) gprintf("<-- %s size=%d\n", __func__, size);
  return size;
}

static INLINE RenID
join_ren(RenID s1, RenID s2) {
  Move p;
  RenID s;

if(0) gprintf("--> %s(%d,%d)\n", __func__, s1, s2);
  assert(s1 != 0);
  assert(s2 != 0);
  assert(s1 != s2);
  assert(libs_ren(s1) > 0);
  assert(libs_ren(s2) > 0);
#if 1
  /* Prefer smaller id for less use of memory. */
  if(s1 > s2)
#else
  /* Collect smaller ren for speed. */
  if(size_ren(s1) < size_ren(s2))
#endif
    { s = s1; s1 = s2; s2 = s; }	/* swap(s1, s2) */
  next_member_ren(last_stone(s1)) = member_ren(s2);
  FOR_ALL_POINT_IN_REN(p, s2) {
    id_ren(p) = s1;
  }
#if USE_ADJACENT_REN
  for(RenID t = 0; t < max_ren_free; t += 1) {
    adjacent_ren[s2][t] = false;
    adjacent_ren[t][s2] = false;
  }
#endif
  size_ren(s1) += collect_ren(s2);
  ADD_DIRTY_IF(s1);
if(0) gprintf("<-- %s ren=%d\n", __func__, s1);
  return s1;
}

static INLINE RenID
add_stone(const Move p, const RenID s) {
if(0) gprintf("<-> %s(%1m,%d)", __func__, p, s);
  assert(s != 0);
  assert(libs_ren(s) > 0);
  size_ren(s) += 1;
  next_member_ren(p) = member_ren(s);
  member_ren(s) = p;
  ADD_DIRTY_IF(s);
if(0) gprintf(" s=%d\n", s);
  return s;
}

/* Find upto MAX_LIBS_REN liberties of a ren <s>.
 * USE mark().
 */
static INLINE int
find_libs_ren(const RenID s) {
#if 0
  Move p, q;
#else
  unsigned int p, q;
#endif
  int n;

if(0) gprintf("--> %s(%d)\n", __func__, s);
  assert(s != 0);
  n = 0;
  reset_mark();
  FOR_ALL_POINT_IN_REN(p, s) {
    FOR_4_ADJACENT_POINT(q, p) {
      if(is_marked(q))		continue;
      mark(q);
      if(!is_empty(q))		continue;
      if(n < MAX_LIBS_REN)	liberty_ren(s)[n] = q;
      n += 1;
    }
  }
  libs_ren(s) = n;
if(0) gprintf("<-- %s libs=%d\n", __func__, n);
  return n;
}

#if 0	/* For debugging */
static volatile Move save_p;
static volatile Color save_c;
#endif

/* Build ren data structure incrementary.
 * Gather several info as well.
 * Returns the ren at <p>.
 */
RenID
update_ren(Move p, Color c) {
  int capture;			/* number of captured stones */
  int libs;
  RenID self = 0;		/* To avoid compiler warning. */
  Move fri, fri1, fri2, fri3;
  Move opp, opp1, opp2, opp3;

if(0) gprintf("--> %s(%1m,%C)\n", __func__, p, c);

#if 0	/* For debugging */
  save_p = p; save_c = c;
#endif
  assert(is_not_ob(p));
  assert(is_empty(p));
  reset_renmark();
  dirty_ren0 = &dirty_ren[0];
  capture = 0;
  libs = 0;

  /* Pack the status into 9 bits and jump! */
  switch(((((((((c - MIN_COLOR)
      << 2) | get_color(NORTH(p))) << 2) | get_color(EAST(p)))
      << 2) | get_color(SOUTH(p))) << 2) | get_color(WEST(p))) {

  /* No friends and no opponents with several liberties (9/256) *
   */
    case 0x000: case 0x100:				/* 4 liberties */
      libs += 1;
    case 0x003: case 0x00c: case 0x030: case 0x0c0:	/* 3 liberties */
    case 0x103: case 0x10c: case 0x130: case 0x1c0:
      libs += 1;
    case 0x00f: case 0x03c: case 0x0f0: case 0x0c3:	/* 2 liberties */
    case 0x10f: case 0x13c: case 0x1f0: case 0x1c3:
      libs += 2;
    /* Create new ren of size 1 and liberties lib. */
      self = create_ren(p, c, libs);
      ADD_DIRTY_IF(self);
      break;

  /* No friends and 1 opponent with several liberties (9+24=33/256) *
   */
    case 0x080: case 0x140:		/* 1 opponent at N, 3 liberties */
    case 0x083: case 0x08c: case 0x0b0:	/* 1 opponent at N, 2 liberties */
    case 0x143: case 0x14c: case 0x170:
    case 0x08f: case 0x0bc: case 0x14f: case 0x17c:	/* 1 liberty */
      opp = NORTH(p);
      goto one_opponent;
    case 0x020: case 0x110:		/* 1 opponent at E, 3 liberties */
    case 0x023: case 0x02c: case 0x0e0:			/* 2 liberties */
    case 0x113: case 0x11c: case 0x1d0:
    case 0x02f: case 0x0e3: case 0x11f: case 0x1d3:	/* 1 liberty */
      opp = EAST(p);
      goto one_opponent;
    case 0x008: case 0x104:		/* 1 opponent at S, 3 liberties */
    case 0x00b: case 0x038: case 0x0c8:			/* 2 liberties */
    case 0x107: case 0x134: case 0x1c4:
    case 0x0f8: case 0x0cb: case 0x1f4: case 0x1c7:	/* 1 liberty */
      opp = SOUTH(p);
      goto one_opponent;
    case 0x002: case 0x101:		/* 1 opponent at W, 3 liberties */
    case 0x00e: case 0x032: case 0x0c2:			/* 2 liberties */
    case 0x10d: case 0x131: case 0x1c1:
    case 0x0f2: case 0x03e: case 0x1f1: case 0x13d:	/* 1 liberty */
      opp = WEST(p);
      goto one_opponent;
one_opponent:
      opp = id_ren(opp);
      CAPTURE(opp);
      self = create_ren(p, c, 0);
      ADD_DIRTY_IF(self);
      break;

  /* No friend and 2 opponents with several liberties (33+22=55/256) *
   */
    case 0x0a0: case 0x150:	/* 2 opponents at N and E with 2 liberties */
    case 0x0ac: case 0x0a3: case 0x15c: case 0x153:	/* 1 liberty */
    case 0x0af: case 0x15f:				/* No liberties */
      opp1 = NORTH(p);
      opp2 = EAST(p);
      goto two_opponents;
    case 0x088: case 0x144:	/* 2 opponents at N and S with 2 liberties */
    case 0x0b8: case 0x08b: case 0x174: case 0x147:	/* 1 liberty */
      opp1 = NORTH(p);
      opp2 = SOUTH(p);
      goto two_opponents;
    case 0x082: case 0x141:	/* 2 opponents at N and W with 2 liberties */
    case 0x0b2: case 0x08e: case 0x171: case 0x14d:	/* 1 liberty */
    case 0x0be: case 0x17d:				/* No liberties */
      opp1 = NORTH(p);
      opp2 = WEST(p);
      goto two_opponents;
    case 0x028: case 0x114:	/* 2 opponents at E and S with 2 liberties */
    case 0x0e8: case 0x02b: case 0x1d4: case 0x117:	/* 1 liberty */
    case 0x0eb: case 0x1d7:				/* No liberties */
      opp1 = EAST(p);
      opp2 = SOUTH(p);
      goto two_opponents;
    case 0x022: case 0x111:	/* 2 opponents at E and W with 2 liberties */
    case 0x0e2: case 0x02e: case 0x1d1: case 0x11d:	/* 1 liberty */
      opp1 = EAST(p);
      opp2 = WEST(p);
      goto two_opponents;
    case 0x00a: case 0x105:	/* 2 opponents at S and W with 2 liberties */
    case 0x0ca: case 0x03a: case 0x1c5: case 0x135:	/* 1 liberty */
    case 0x0fa: case 0x1f5:				/* No liberties */
      opp1 = SOUTH(p);
      opp2 = WEST(p);
      goto two_opponents;
two_opponents:
      opp1 = id_ren(opp1);
      opp2 = id_ren(opp2);
      if(opp1 == opp2) {
      	CAPTURE(opp1);
      } else {
      	CAPTURE(opp1);
      	CAPTURE(opp2);
      }
      self = create_ren(p, c, 0);
      ADD_DIRTY_IF(self);
      break;

  /* No friends and 3 opponents with several liberties (55+8=63/256) *
   */
    case 0x0a8: case 0x154:	/* 3 opponents at N, E, S with 1 liberty */
    case 0x0ab: case 0x157:				/* No liberty */
      opp1 = NORTH(p);
      opp2 = EAST(p);
      opp3 = SOUTH(p);
      goto three_opponents;
    case 0x0a2: case 0x151:	/* 3 opponents at N, E, W with 1 liberty */
    case 0x0ae: case 0x15d:				/* No liberty */
      opp1 = NORTH(p);
      opp2 = EAST(p);
      opp3 = WEST(p);
      goto three_opponents;
    case 0x08a: case 0x145:	/* 3 opponents at N, S, W with 1 liberty */
    case 0x0ba: case 0x175:				/* No liberty */
      opp1 = NORTH(p);
      opp2 = SOUTH(p);
      opp3 = WEST(p);
      goto three_opponents;
    case 0x02a: case 0x115:	/* 3 opponents at E, S, W with 1 liberty */
    case 0x0ea: case 0x1d5:				/* No liberty */
      opp1 = EAST(p);
      opp2 = SOUTH(p);
      opp3 = WEST(p);
      goto three_opponents;
three_opponents:
      CAPTURE3(opp1, opp2, opp3);
      self = create_ren(p, c, 0);
      ADD_DIRTY_IF(self);
      break;

  /* No friends and 4 opponents with no liberties (63+1=64/256) *
   */
    case 0x0aa: case 0x155:				/* 4 opponents */
      /* Collect unique adjacent rens and liberties. */
#if 0
      opp1 = NORTH(p);
      assert(get_color(opp1) == opposit_color(c));
      opp1 = id_ren(opp1);
      CAPTURE(opp1);
      opp1 = EAST(p); 
      if(get_color(opp1) == opposit_color(c)) {
	opp1 = id_ren(opp1);
	if(!is_marked_ren(opp1))	CAPTURE(opp1);
      }
      opp1 = SOUTH(p); 
      if(get_color(opp1) == opposit_color(c)) {
	opp1 = id_ren(opp1);
	if(!is_marked_ren(opp1))	CAPTURE(opp1);
      }
      opp1 = WEST(p); 
      if(get_color(opp1) == opposit_color(c)) {
	opp1 = id_ren(opp1);
	if(!is_marked_ren(opp1))	CAPTURE(opp1);
      }
#elif 0
      opp1 = id_ren(NORTH(p));
      opp2 = id_ren(EAST(p)); 
      opp3 = id_ren(SOUTH(p)); 
      opp = id_ren(WEST(p)); 
      if(opp1 == opp2) {
      	if(opp1 == opp3) {	/* opp1 = opp2 = opp3 */
      	  if(opp1 == opp) {
	    CAPTURE(opp1);	/* opp1 = opp2 = opp3 = opp */
	  } else {
	    CAPTURE(opp1);	/* opp1 = opp2 = opp3 != opp */
	    CAPTURE(opp);
	  }
	} else {		/* opp1 = opp2 != opp3 */
	  if(opp1 == opp) {	/* opp1 = opp2 = opp != opp3 */
	    CAPTURE(opp1);
	    CAPTURE(opp3);
	  } else {		/* opp1 = opp2 != opp3, opp1 != opp */
	    if(opp3 == opp) {	/* opp1 = opp2, opp3 = opp */
	      CAPTURE(opp1);
	      CAPTURE(opp3);
	    } else {		/* opp1 = opp2, opp3, opp */
	      CAPTURE(opp1);
	      CAPTURE(opp3);
	      CAPTURE(opp);
	    }
	  }
	}
      } else {			/* opp1 != opp2 */
	if(opp1 == opp3) {	/* opp1 = opp3 != opp2 */
	  if(opp1 == opp) {	/* opp1 = opp3 = opp, opp2 */
	    CAPTURE(opp1);
	    CAPTURE(opp2);
	  } else {		/* opp1 = opp3 != opp2, != opp */
	    if(opp2 == opp) {	/* opp1 = opp3, opp2 = opp */
	      CAPTURE(opp1);
	      CAPTURE(opp2);
	    } else {		/* opp1 = opp3, opp2, opp */
	      CAPTURE(opp1);
	      CAPTURE(opp2);
	      CAPTURE(opp);
	    }
	  }
	} else {		/* opp1 != opp2, opp1 != opp3 */
	  if(opp2 == opp3) {	/* opp1, opp2 = opp3 */
	    if(opp2 == opp) {	/* opp1, opp2 = opp3 = opp */
	      CAPTURE(opp1);
	      CAPTURE(opp2);
	    } else {
	      if(opp1 == opp) { /* opp1 = opp, opp2 = opp3 */
		CAPTURE(opp1);
		CAPTURE(opp2);
	      } else {		/* opp1, opp2 = opp3, opp */
		CAPTURE(opp1);
		CAPTURE(opp2);
		CAPTURE(opp);
	      }
	    }
	  } else {		/* opp1 != opp2, opp1 != opp3, opp2 != opp3 */
	    if(opp == opp1) {	/* opp1 = opp, opp2, opp3 */
	      CAPTURE(opp1);
	      CAPTURE(opp2);
	      CAPTURE(opp3);
	    } else {
	      CAPTURE(opp1);
	      CAPTURE(opp2);
	      CAPTURE(opp3);
	      CAPTURE(opp);
	    }
	  }
	}
      }
#else
      /* Collect unique adjacent rens and liberties. */
      {
      	int k, i;
      	Move q;
      	RenID s, adj_ren[4 + 1];

	k = 0;
	FOR_4_ADJACENT_POINT(q, p) {
	  assert(get_color(q) == opposit_color(c));
	  s = id_ren(q);
	  assert(s != 0);
	  for(i = 0; i < k; i++) {
	    if(s == adj_ren[i]) {
	      break;
	    }
	  }
	  if(i >= k) {	/* new */
	    adj_ren[k] = s;
	    k++;
	  }
	}
	for(i = 0; i < k; i++) {
	  s = adj_ren[i];
	  CAPTURE(s);
	}
      }
#endif
      self = create_ren(p, c, 0);
      ADD_DIRTY_IF(self);
      break;

  /* 1 friend and no opponents (64+24=88/256) with several liberties *
   */
    case 0x040: case 0x180:		/* 1 friend at N with 3 liberties */
    case 0x043: case 0x04c: case 0x070:			/* with 2 liberties */
    case 0x183: case 0x18c: case 0x1b0:
    case 0x04f: case 0x07c: case 0x18f: case 0x1bc:	/* with 1 liberty */
      fri = NORTH(p);
      goto merge;
    case 0x010: case 0x120:		/* 1 friend at E with 3 liberties */
    case 0x013: case 0x01c: case 0x0d0:			/* with 2 liberties */
    case 0x123: case 0x12c: case 0x1e0:
    case 0x01f: case 0x0d3: case 0x12f: case 0x1e3:	/* with 1 liberty */
      fri = EAST(p);
      goto merge;
    case 0x004: case 0x108:		/* 1 friend at S with 3 liberties */
    case 0x007: case 0x034: case 0x0c4:			/* with 2 liberties */
    case 0x10b: case 0x138: case 0x1c8:
    case 0x0f4: case 0x0c7: case 0x1f8: case 0x1cb:	/* with 1 liberty */
      fri = SOUTH(p);
      goto merge;
    case 0x001: case 0x102:		/* 1 friend at W with 3 liberties */
    case 0x00d: case 0x031: case 0x0c1:			/* with 2 liberties */
    case 0x10e: case 0x132: case 0x1c2:
    case 0x0f1: case 0x03d: case 0x1f2: case 0x13e:	/* with 1 liberty */
      fri = WEST(p);
      goto merge;
merge:
      self = add_stone(p, id_ren(fri));
      break;

  /* 1 friend and 1 opponent (88+44=132/256) *
   */
    case 0x060: case 0x190:		/* friend at N, opp at E, 2 libs */
    case 0x063: case 0x06c: case 0x193: case 0x19c:	/* 1 lib */
    case 0x06f: case 0x19f:				/* No libs */
      fri = NORTH(p);
      opp = EAST(p);
      goto one_and_one;
    case 0x048: case 0x184:		/* friend at N, opp at S, 2 libs */
    case 0x078: case 0x04b: case 0x1b4: case 0x187:	/* 1 lib */
      fri = NORTH(p);
      opp = SOUTH(p);
      goto one_and_one;
    case 0x042: case 0x181:		/* friend at N, opp at W, 2 libs */
    case 0x072: case 0x04e: case 0x1b1: case 0x18d:	/* 1 lib */
    case 0x07e: case 0x1bd:				/* No libs */
      fri = NORTH(p);
      opp = WEST(p);
      goto one_and_one;
    case 0x090: case 0x160:		/* friend at E, opp at N, 2 libs */
    case 0x093: case 0x09c: case 0x163: case 0x16c:	/* 1 lib */
    case 0x09f: case 0x16f:				/* No libs */
      fri = EAST(p);
      opp = NORTH(p);
      goto one_and_one;
    case 0x018: case 0x124:		/* friend at E, opp at S, 2 libs */
    case 0x0d8: case 0x01b: case 0x1e4: case 0x127:	/* 1 lib */
    case 0x0db: case 0x1e7:				/* No libs */
      fri = EAST(p);
      opp = SOUTH(p);
      goto one_and_one;
    case 0x012: case 0x121:		/* friend at E, opp at W, 2 libs */
    case 0x0d2: case 0x01e: case 0x1e1: case 0x12d:	/* 1 lib */
      fri = EAST(p);
      opp = WEST(p);
      goto one_and_one;
    case 0x084: case 0x148:		/* friend at S, opp at N, 2 libs */
    case 0x087: case 0x0b4: case 0x178: case 0x14b:	/* 1 lib */
      fri = SOUTH(p);
      opp = NORTH(p);
      goto one_and_one;
    case 0x024: case 0x118:		/* friend at S, opp at E, 2 libs */
    case 0x0e4: case 0x027: case 0x1d8: case 0x11b:	/* 1 lib */
    case 0x0e7: case 0x1db:				/* No libs */
      fri = SOUTH(p);
      opp = EAST(p);
      goto one_and_one;
    case 0x006: case 0x109:		/* friend at S, opp at W, 2 libs */
    case 0x0c6: case 0x036: case 0x1c9: case 0x139:	/* 1 lib */
    case 0x0f6: case 0x1f9:				/* No libs */
      fri = SOUTH(p);
      opp = WEST(p);
      goto one_and_one;
    case 0x081: case 0x142:		/* friend at W, opp at N, 2 libs */
    case 0x0b1: case 0x08d: case 0x172: case 0x14e:	/* 1 lib */
    case 0x0bd: case 0x17e:				/* No libs */
      fri = WEST(p);
      opp = NORTH(p);
      goto one_and_one;
    case 0x021: case 0x112:		/* friend at W, opp at E, 2 libs */
    case 0x0e1: case 0x02d: case 0x1d2: case 0x11e:	/* 1 lib */
      fri = WEST(p);
      opp = EAST(p);
      goto one_and_one;
    case 0x009: case 0x106:		/* friend at W, opp at S, 2 libs */
    case 0x0c9: case 0x039: case 0x1c6: case 0x136:	/* 1 lib */
    case 0x0f9: case 0x1f6:				/* No libs */
      fri = WEST(p);
      opp = SOUTH(p);
      goto one_and_one;
one_and_one:
      opp = id_ren(opp);
      CAPTURE(opp);
      self = add_stone(p, id_ren(fri));
      break;

  /* 1 friend and 2 opponents (132+24=156/256) *
   */
    case 0x068: case 0x194:	/* 1 friend at N, 2 opps at E, S, 1 lib */
    case 0x06b: case 0x197:				/* No liberties */
      fri = NORTH(p);
      opp1 = EAST(p);
      opp2 = SOUTH(p);
      goto one_and_two;
    case 0x062: case 0x191:	/* 1 friend at N, 2 opps at E, W, 1 lib */
    case 0x06e: case 0x19d:				/* No liberties */
      fri = NORTH(p);
      opp1 = EAST(p);
      opp2 = WEST(p);
      goto one_and_two;
    case 0x04a: case 0x185:	/* 1 friend at N, 2 opps at S, W, 1 lib */
    case 0x07a: case 0x1b5:				/* No liberties */
      fri = NORTH(p);
      opp1 = SOUTH(p);
      opp2 = WEST(p);
      goto one_and_two;
    case 0x098: case 0x164:	/* 1 friend at E, 2 opps at N, S, 1 lib */
    case 0x09b: case 0x167:				/* No liberties */
      fri = EAST(p);
      opp1 = NORTH(p);
      opp2 = SOUTH(p);
      goto one_and_two;
    case 0x092: case 0x161:	/* 1 friend at E, 2 opps at N, W, 1 lib */
    case 0x09e: case 0x16d:				/* No liberties */
      fri = EAST(p);
      opp1 = NORTH(p);
      opp2 = WEST(p);
      goto one_and_two;
    case 0x01a: case 0x125:	/* 1 friend at E, 2 opps at S, W, 1 lib */
    case 0x0da: case 0x1e5:				/* No liberties */
      fri = EAST(p);
      opp1 = SOUTH(p);
      opp2 = WEST(p);
      goto one_and_two;
    case 0x0a4: case 0x158:	/* 1 friend at S, 2 opps at N, E, 1 lib */
    case 0x0a7: case 0x15b:				/* No liberties */
      fri = SOUTH(p);
      opp1 = NORTH(p);
      opp2 = EAST(p);
      goto one_and_two;
    case 0x086: case 0x149:	/* 1 friend at S, 2 opps at N, W, 1 lib */
    case 0x0b6: case 0x179:				/* No liberties */
      fri = SOUTH(p);
      opp1 = NORTH(p);
      opp2 = WEST(p);
      goto one_and_two;
    case 0x026: case 0x119:	/* 1 friend at S, 2 opps at E, W, 1 lib */
    case 0x0e6: case 0x1d9:				/* No liberties */
      fri = SOUTH(p);
      opp1 = EAST(p);
      opp2 = WEST(p);
      goto one_and_two;
    case 0x0a1: case 0x152:	/* 1 friend at W, 2 opps at N, E, 1 lib */
    case 0x0ad: case 0x15e:				/* No liberties */
      fri = WEST(p);
      opp1 = NORTH(p);
      opp2 = EAST(p);
      goto one_and_two;
    case 0x089: case 0x146:	/* 1 friend at W, 2 opps at N, S, 1 lib */
    case 0x0b9: case 0x176:				/* No liberties */
      fri = WEST(p);
      opp1 = NORTH(p);
      opp2 = SOUTH(p);
      goto one_and_two;
    case 0x029: case 0x116:	/* 1 friend at W, 2 opps at E, S, 1 lib */
    case 0x0e9: case 0x1d6:				/* No liberties */
      fri = WEST(p);
      opp1 = EAST(p);
      opp2 = SOUTH(p);
      goto one_and_two;
one_and_two:
      opp1 = id_ren(opp1);
      opp2 = id_ren(opp2);
      if(opp1 == opp2) {
      	CAPTURE(opp1);
      } else {
      	CAPTURE(opp1);
      	CAPTURE(opp2);
      }
      self = add_stone(p, id_ren(fri));
      break;

  /* 1 friend and 3 opponents (156+4=160/256) *
   */
    case 0x06a: case 0x195:	/* 1 friend at N, 3 opps at E, S, W */
      fri = NORTH(p);
      opp1 = EAST(p);
      opp2 = SOUTH(p);
      opp3 = WEST(p);
      goto one_and_three;
    case 0x09a: case 0x165:	/* 1 friend at E, 3 opps at N, S, W */
      fri = EAST(p);
      opp1 = NORTH(p);
      opp2 = SOUTH(p);
      opp3 = WEST(p);
      goto one_and_three;
    case 0x0a6: case 0x159:	/* 1 friend at S, 3 opps at N, E, W */
      fri = SOUTH(p);
      opp1 = NORTH(p);
      opp2 = EAST(p);
      opp3 = WEST(p);
      goto one_and_three;
    case 0x0a9: case 0x156:	/* 1 friend at W, 3 opps at N, E, S */
      fri = WEST(p);
      opp1 = NORTH(p);
      opp2 = EAST(p);
      opp3 = SOUTH(p);
      goto one_and_three;
one_and_three:
      CAPTURE3(opp1, opp2, opp3);
      self = add_stone(p, id_ren(fri));
      break;

  /* 2 friends and no opponents (160+22=182/256) *
   */
    case 0x050: case 0x1a0:	/* 2 friends at N and E, 2 liberties */
    case 0x053: case 0x05c: case 0x1a3: case 0x1ac:	/* 1 liberty */
    case 0x05f: case 0x1af:				/* No liberty */
      fri1 = NORTH(p);
      fri2 = EAST(p);
      goto two_friends;
    case 0x044: case 0x188:	/* 2 friends at N and S, 2 liberties */
    case 0x074: case 0x047: case 0x1b8: case 0x18b:	/* 1 liberty */
      fri1 = NORTH(p);
      fri2 = SOUTH(p);
      goto two_friends;
    case 0x041: case 0x182:	/* 2 friends at N and W, 2 liberties */
    case 0x071: case 0x04d: case 0x1b2: case 0x18e:	/* 1 liberty */
    case 0x07d: case 0x1be:				/* No liberties */
      fri1 = NORTH(p);
      fri2 = WEST(p);
      goto two_friends;
    case 0x014: case 0x128:	/* 2 friends at E and S, 2 liberties */
    case 0x0d4: case 0x017: case 0x12b: case 0x1e8:	/* 1 liberty */
    case 0x0d7: case 0x1eb:				/* No liberties */
      fri1 = EAST(p);
      fri2 = SOUTH(p);
      goto two_friends;
    case 0x011: case 0x122:	/* 2 friends at E and W, 2 liberties */
    case 0x0d1: case 0x01d: case 0x1e2: case 0x12e:	/* 1 liberty */
      fri1 = EAST(p);
      fri2 = WEST(p);
      goto two_friends;
    case 0x005: case 0x10a:	/* 2 friends at S and W, 2 liberties */
    case 0x035: case 0x0c5: case 0x13a: case 0x1ca:	/* 1 liberty */
    case 0x0f5: case 0x1fa:				/* No liberties */
      fri1 = SOUTH(p);
      fri2 = WEST(p);
      goto two_friends;

  /* 2 friends and 1 opponent (182+24=206/256) *
   */
    case 0x058: case 0x1a4:	/* 2 friends at N, E and 1 opp at S, 1 lib */
    case 0x05b: case 0x1a7:				/* No liberties */
      fri1 = NORTH(p);
      fri2 = EAST(p);
      opp = SOUTH(p);
      goto two_and_one;
    case 0x052: case 0x1a1:	/* 2 friends at N, E and 1 opp at W, 1 lib */
    case 0x05e: case 0x1ad:				/* No liberties */
      fri1 = NORTH(p);
      fri2 = EAST(p);
      opp = WEST(p);
      goto two_and_one;
    case 0x064: case 0x198:	/* 2 friends at N, S and 1 opp at E, 1 lib */
    case 0x067: case 0x19b:				/* No liberties */
      fri1 = NORTH(p);
      fri2 = SOUTH(p);
      opp = EAST(p);
      goto two_and_one;
    case 0x046: case 0x189:	/* 2 friends at N, S and 1 opp at W, 1 lib */
    case 0x076: case 0x1b9:				/* No liberties */
      fri1 = NORTH(p);
      fri2 = SOUTH(p);
      opp = WEST(p);
      goto two_and_one;
    case 0x061: case 0x192:	/* 2 friends at N, W and 1 opp at E, 1 lib */
    case 0x06d: case 0x19e:				/* No liberties */
      fri1 = NORTH(p);
      fri2 = WEST(p);
      opp = EAST(p);
      goto two_and_one;
    case 0x049: case 0x186:	/* 2 friends at N, W and 1 opp at S, 1 lib */
    case 0x079: case 0x1b6:				/* No liberties */
      fri1 = NORTH(p);
      fri2 = WEST(p);
      opp = SOUTH(p);
      goto two_and_one;
    case 0x094: case 0x168:	/* 2 friends at E, S and 1 opp at N, 1 lib */
    case 0x097: case 0x16b:				/* No liberties */
      fri1 = EAST(p);
      fri2 = SOUTH(p);
      opp = NORTH(p);
      goto two_and_one;
    case 0x016: case 0x129:	/* 2 friends at E, S and 1 opp at W, 1 lib */
    case 0x0d6: case 0x1e9:				/* No liberties */
      fri1 = EAST(p);
      fri2 = SOUTH(p);
      opp = WEST(p);
      goto two_and_one;
    case 0x091: case 0x162:	/* 2 friends at E, W and 1 opp at N, 1 lib */
    case 0x09d: case 0x16e:				/* No liberties */
      fri1 = EAST(p);
      fri2 = WEST(p);
      opp = NORTH(p);
      goto two_and_one;
    case 0x019: case 0x126:	/* 2 friends at E, W and 1 opp at S, 1 lib */
    case 0x0d9: case 0x1e6:				/* No liberties */
      fri1 = EAST(p);
      fri2 = WEST(p);
      opp = SOUTH(p);
      goto two_and_one;
    case 0x085: case 0x14a:	/* 2 friends at S, W and 1 opp at N, 1 lib */
    case 0x0b5: case 0x17a:				/* No liberties */
      fri1 = SOUTH(p);
      fri2 = WEST(p);
      opp = NORTH(p);
      goto two_and_one;
    case 0x025: case 0x11a:	/* 2 friends at S, W and 1 opp at E, 1 lib */
    case 0x0e5: case 0x1da:				/* No liberties */
      fri1 = SOUTH(p);
      fri2 = WEST(p);
      opp = EAST(p);
      goto two_and_one;
two_and_one:
      opp = id_ren(opp);
      CAPTURE(opp);
      goto two_friends;

  /* 2 friends and 2 opponents (206+6=212/256) *
   */
    case 0x05a: case 0x1a5:	/* 2 friends at N, E and 2 opp at S, W */
      fri1 = NORTH(p);
      fri2 = EAST(p);
      opp1 = SOUTH(p);
      opp2 = WEST(p);
      goto two_and_two;
    case 0x066: case 0x199:	/* 2 friends at N, S and 2 opp at E, W */
      fri1 = NORTH(p);
      fri2 = SOUTH(p);
      opp1 = EAST(p);
      opp2 = WEST(p);
      goto two_and_two;
    case 0x069: case 0x196:	/* 2 friends at N, W and 2 opp at E, S */
      fri1 = NORTH(p);
      fri2 = WEST(p);
      opp1 = EAST(p);
      opp2 = SOUTH(p);
      goto two_and_two;
    case 0x096: case 0x169:	/* 2 friends at E, S and 2 opp at N, W */
      fri1 = EAST(p);
      fri2 = SOUTH(p);
      opp1 = NORTH(p);
      opp2 = WEST(p);
      goto two_and_two;
    case 0x099: case 0x166:	/* 2 friends at E, W and 2 opp at N, S */
      fri1 = EAST(p);
      fri2 = WEST(p);
      opp1 = NORTH(p);
      opp2 = SOUTH(p);
      goto two_and_two;
    case 0x0a5: case 0x15a:	/* 2 friends at S, W and 2 opp at N, E */
      fri1 = SOUTH(p);
      fri2 = WEST(p);
      opp1 = NORTH(p);
      opp2 = EAST(p);
      goto two_and_two;
two_and_two:
      opp1 = id_ren(opp1);
      opp2 = id_ren(opp2);
      if(opp1 == opp2) {
      	CAPTURE(opp1);
      } else {
      	CAPTURE(opp1);
      	CAPTURE(opp2);
      }
      goto two_friends;
two_friends:
      fri1 = id_ren(fri1);
      fri2 = id_ren(fri2);
      if(fri1 != fri2) {
	fri1 = join_ren(fri1, fri2);
      }
      self = add_stone(p, fri1);
      break;

  /* 3 friends and no opponents (212+8=220/256) *
   */
    case 0x054: case 0x1a8:	/* 3 friends at N, E and S, 1 liberty */
    case 0x057: case 0x1ab:				/* No liberty */
      fri1 = NORTH(p);
      fri2 = EAST(p);
      fri3 = SOUTH(p);
      goto three_friends;
    case 0x051: case 0x1a2:	/* 3 friends at N, E and W, 1 liberty */
    case 0x05d: case 0x1ae:				/* No liberty */
      fri1 = NORTH(p);
      fri2 = EAST(p);
      fri3 = WEST(p);
      goto three_friends;
    case 0x045: case 0x18a:	/* 3 friends at N, S and W, 1 liberty */
    case 0x075: case 0x1ba:				/* No liberty */
      fri1 = NORTH(p);
      fri2 = SOUTH(p);
      fri3 = WEST(p);
      goto three_friends;
    case 0x015: case 0x12a:	/* 3 friends at E, S and W, 1 liberty */
    case 0x0d5: case 0x1ea:				/* No liberty */
      fri1 = EAST(p);
      fri2 = SOUTH(p);
      fri3 = WEST(p);
      goto three_friends;

  /* 3 friends and 1 opponent (220+4=224/256) *
   */
    case 0x056: case 0x1a9:	/* 3 friends at N, E, S and 1 opp at W */
      fri1 = NORTH(p);
      fri2 = EAST(p);
      fri3 = SOUTH(p);
      opp = WEST(p);
      goto three_and_one;
    case 0x059: case 0x1a6:	/* 3 friends at N, E, W and 1 opp at S */
      fri1 = NORTH(p);
      fri2 = EAST(p);
      fri3 = WEST(p);
      opp = SOUTH(p);
      goto three_and_one;
    case 0x065: case 0x19a:	/* 3 friends at N, S, W and 1 opp at E */
      fri1 = NORTH(p);
      fri2 = SOUTH(p);
      fri3 = WEST(p);
      opp = EAST(p);
      goto three_and_one;
    case 0x095: case 0x16a:	/* 3 friends at E, S, W and 1 opp at N */
      fri1 = EAST(p);
      fri2 = SOUTH(p);
      fri3 = WEST(p);
      opp = NORTH(p);
      goto three_and_one;
three_and_one:
      opp = id_ren(opp);
      CAPTURE(opp);
three_friends:
      fri1 = id_ren(fri1);
      fri2 = id_ren(fri2);
      fri3 = id_ren(fri3);
      switch((((fri1 != fri2) << 1) | (fri1 != fri3)) << 1 | (fri2 != fri3)) {
	case 1: case 2: case 4:	/* Impossible cases */
	  FATAL("merging three friends %d %d %d\n", fri1, fri2, fri3);
	  /* Never return here. */
	case 7:	/* All are different */
	  fri1 = join_ren(fri1, fri2);
	  /* Fall through */
	case 3:	/* fri1 == fri2 */
	  fri2 = fri3;
	  /* Fall through */ /* join(fri1, fri3); */
	case 5:	/* fri1 == fri3 */
	case 6:	/* fri2 == fri3 */
	  fri1 = join_ren(fri1, fri2);
	  /* Fall through */
	case 0:	/* All 3 are the same */
	break;
      }
      self = add_stone(p, fri1);
      break;

  /* 4 friends (224+1=225/256) *
   */
    case 0x055: case 0x1aa:	/* 4 friends at N, E, S and W */
#if 0
      fri = id_ren(NORTH(p));
      self = add_stone(p, fri);
      fri = id_ren(EAST(p));
      if(fri != self)	self = join_ren(self, fri);
      fri = id_ren(SOUTH(p));
      if(fri != self)	self = join_ren(self, fri);
      fri = id_ren(WEST(p));
      if(fri != self)	self = join_ren(self, fri);
#else
      fri1 = NORTH(p);
      fri1 = id_ren(fri1);
      self = add_stone(p, fri1);
      fri2 = EAST(p);
      fri2 = id_ren(fri2);
      if(fri2 != self) {
	self = join_ren(self, fri2);
      }
      fri3 = SOUTH(p);
      fri3 = id_ren(fri3);
      if(fri3 != self) {
	self = join_ren(self, fri3);
      }
      fri = WEST(p);
      fri = id_ren(fri);
      if(fri != self) {
	self = join_ren(self, fri);
      }
#endif
      break;

  /* Impossible cases (225+31=256/256) *
   */
    case 0x0ff: case 0x1ff:				/* 4 exteriors */
    case 0x03f: case 0x0fc: case 0x0f3: case 0x0cf:	/* 3 exteriors */
    case 0x13f: case 0x1fc: case 0x1f3: case 0x1cf:
    case 0x07f: case 0x0fd: case 0x0f7: case 0x0df:	/* + 1 white */
    case 0x17f: case 0x1fd: case 0x1f7: case 0x1df:
    case 0x0bf: case 0x0fe: case 0x0fb: case 0x0ef:	/* + 1 black */
    case 0x1bf: case 0x1fe: case 0x1fb: case 0x1ef:
    case 0x033: case 0x0cc:		/* 2 exteriors in face to face */
    case 0x133: case 0x1cc:
    case 0x037: case 0x073: case 0x0cd:	case 0x0dc:	/* + 1 white */
    case 0x137: case 0x173: case 0x1cd:	case 0x1dc:
    case 0x03b: case 0x0b3: case 0x0ce:	case 0x0ec:	/* + 1 black */
    case 0x13b: case 0x1b3: case 0x1ce:	case 0x1ec:
    case 0x077: case 0x0dd:				/* + 2 whites */
    case 0x177: case 0x1dd:
    case 0x0bb: case 0x0ee:				/* + 2 blacks */
    case 0x1bb: case 0x1ee:
    case 0x07b: case 0x0b7: case 0x0ed: case 0x0de:	/* + white and black */
    case 0x17b: case 0x1b7: case 0x1ed: case 0x1de:
      FATAL("Impossible pattern 0x%0x at %1m\n", ((((((((c - MIN_COLOR)
	  << 2) | get_color(NORTH(p))) << 2) | get_color(EAST(p)))
	  << 2) | get_color(SOUTH(p))) << 2) | get_color(WEST(p)), p);
    break;
  }

  *dirty_ren0 = 0;
  captured_stones = capture;

  set_color(p, c);
#if USE_EMPTY_LIST
  remove_from_empty_list(p);
#endif
  id_ren(p) = self;

  /* Update liberties of dirty rens. */
  for(RenID *x = &dirty_ren[0]; *x != 0; x++) {
    int libs;
    RenID s = *x;
    if(color_ren(s) == EMPTY)	continue;
    assert(s != 0);
    assert((s == self) || libs_ren(s) > 0);
    libs = find_libs_ren(s);
#if 0
    assert(libs > 0);
#else
    if(libs <= 0) FATAL("p=%d c=%C self=%d s=%d x=%p\n", p, c, self, s, x);
#endif
if(0) gprintf("- update_libs ren=%d libs=%d\n", *x, libs_ren(*x));
  }

if(0) gprintf("<-- %s ren=%d libs=%d\n", __func__, self, libs_ren(self));
  return self;
}
